/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.animal;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

import de.unibi.techfak.isy.lppba.descriptors.ExactAge;
import de.unibi.techfak.isy.lppba.descriptors.Time;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil;

/**
 * This annotates more specific description of a lab animals age based on a
 * specific time like "27 years old" or "2 weeks of age".
 * 
 * This Annotator depends on the TimeAnnotator.
 * 
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ExactAgeAnnotator extends JCasAnnotator_ImplBase {

	public static final Pattern KEYWORDS = Pattern.compile("age|old");
	/**
	 * This decides whether a word is an age specifier as given by the Keywords
	 * pattern above.
	 */
	public static final UIMAAnnotationCompositionUtil.UIMASemanticProbabilityModel<Token> MODEL = new UIMAAnnotationCompositionUtil.UIMASemanticProbabilityModel<Token>() {
		@Override
		public double calculateSemanticProbability(Token candidate) {
			final String coveredText = candidate.getCoveredText().toLowerCase();
			final Matcher matcher = KEYWORDS.matcher(coveredText);
			if (matcher.matches()) {
				return 1;
			} else {
				return 0;
			}
		}
	};
	/**
	 * We expect an age specifier roughly in 10 characters distance.
	 */
	public static final double SIGMA = 10;
	/**
	 * The threshold we require for the probability of an age specifier.
	 */
	public static final double THRESHOLD = 0;
	/**
	 * The maximum age we accept in seconds. (200 years)
	 */
	public static final double MAXAGE = 200 * 365 * 24 * 60 * 60;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final Collection<Time> times = JCasUtil.select(jcas, Time.class);
		for (final Time time : times) {
			// times below 0 and above MAXAGE are regarded as impossible.
			if (time.getSIValue() != -1
					&& (time.getSIValue() < 0 || time.getSIValue() > MAXAGE)) {
				continue;
			}

			// try to find an age specifier.
			final ProbableHypothesis<UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate<Token>> mapWrapper = UIMAAnnotationCompositionUtil
					.calculateMaxProbHypothesis(time, MODEL, Token.class, jcas, SIGMA);
			final UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate<Token> map = mapWrapper
					.getCandidate();
			if (map != null && mapWrapper.getProbability() > THRESHOLD) {
				// if we found one, construct the annotation.
				final int begin;
				if (time.getBegin() < map.getBegin()) {
					begin = time.getBegin();
				} else {
					begin = map.getBegin();
				}
				final int end;
				if (time.getEnd() > map.getEnd()) {
					end = time.getEnd();
				} else {
					end = map.getEnd();
				}

				final ExactAge age = new ExactAge(jcas, begin, end);
				age.setTime(time);
				age.setAnnotationId(AnnotationCounter.getUniqueId());
				age.addToIndexes();
			}
		}
	}
}
