/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.animal;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.unibi.techfak.isy.lppba.descriptors.Age;
import de.unibi.techfak.isy.lppba.descriptors.ExactAge;
import de.unibi.techfak.isy.lppba.descriptors.Integer;
import de.unibi.techfak.isy.lppba.descriptors.LabGroup;
import de.unibi.techfak.isy.lppba.descriptors.SpecialLaboratoryAnimal;
import de.unibi.techfak.isy.lppba.descriptors.Organism;
import de.unibi.techfak.isy.lppba.descriptors.Quantity;
import de.unibi.techfak.isy.lppba.descriptors.Range;
import de.unibi.techfak.isy.lppba.descriptors.Gender;
import de.unibi.techfak.isy.lppba.descriptors.Weight;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMASemanticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMACompositor;
import de.unibi.techfak.scie.probability_model.uima.UIMAWordSyntacticProbabilityModel;
import de.unibi.techfak.scie.util.CachedJCasUtil;

/**
 * This Annotator accumulates the output of all other Taxonomy-Annotators to a Lab Group Annotation.
 *
 * It iterates over all organism annotations and searches in the current sentence as well as the
 * previous and the following one for additional information, namely:
 *
 * <ol>
 * <li>The species of lab animals used (OrganismAnnotator and LaboratoryRatAnnotator)</li>
 * <li>The age of the lab animals (AgeAnotator)</li>
 * <li>The gender of the lab anomals (GenderAnnotator)</li>
 * <li>The number of lab animals in the group (NumberAnnotator and WordNumberAnnotator).</li>
 * </ol>
 *
 * Please note that all forementioned Annotators have to be executed before the LabGroupAnnotator to
 * ensure meaningful results.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class LabGroupAnnotator extends org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public LabGroupAnnotator() {
	}
	public static final AgeModel AGEMODEL = new AgeModel();
	public static final GenderModel GENDERMODEL = new GenderModel();
	public static final WeightModel WEIGHTMODEL = new WeightModel();
	public static final NumberModel NUMBERMODEL = new NumberModel();
	public static final UIMASemanticProbabilityModel[] MODELS = {AGEMODEL,
		GENDERMODEL, WEIGHTMODEL};
	public static final Class[] CLASSES = {Age.class, Gender.class, Weight.class};
	public static final LabGroupCompositor COMPOSITOR = new LabGroupCompositor();
	/**
	 * This specifies about how far we search for annotations that belong to a lab group. For more
	 * details on this topic have a look at the AnnotationCompositionUtil class.
	 */
	public static final double SIGMA = 6;
	/**
	 * LabGroup annotations below this probability threshold are thrown away. This threshold was
	 * derived by using statistics over the available 20 test papers.
	 */
	public static final double THRESHOLD = 0.2;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final UIMAWordSyntacticProbabilityModel syntacticModel =
				new UIMAWordSyntacticProbabilityModel(
				SIGMA, jcas);
		COMPOSITOR.setSyntacticModel(syntacticModel);
		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		final Collection<Organism> organisms = jcasUtil.select(Organism.class);
		for (final Organism orga : organisms) {
			// look if there is another organism annotation that covers this one. If
			// so the other one
			// is more specific and we disregard this one.
			final List<Organism> coveringOrganisms = jcasUtil.selectCovering(
					Organism.class, orga.getBegin(), orga.getEnd());
			boolean existsCovering = false;
			for (final Organism coveringOrga : coveringOrganisms) {
				if (coveringOrga.getBegin() < orga.getBegin() ||
						 coveringOrga.getEnd() > orga.getEnd()) {
					existsCovering = true;
					break;
				}
			}
			if (existsCovering) {
				continue;
			}
			// now use the CompositionUtil to create every possible combination
			// of annotations with the sorrounding age, gender and weight annotations.
			UIMAAnnotationCompositionUtil.calculateComposita(orga, syntacticModel,
					MODELS, CLASSES, COMPOSITOR, jcas);
		}
	}

	/**
	 * The likelihood model for ages.
	 */
	public static class AgeModel extends UIMASemanticProbabilityModel<Age> {

		public AgeModel() {
		}

		@Override
		public double calculateSemanticProbability(Age candidate) {
			// every Age is equally probable.
			return 1;
		}
	}

	/**
	 * The trivial likelihood Model for Gender.
	 */
	public static class GenderModel extends UIMASemanticProbabilityModel<Gender> {

		public GenderModel() {
		}

		@Override
		public double calculateSemanticProbability(Gender candidate) {
			// every gender is equally probable.
			return 1;
		}
	}

	/**
	 * The likelihood model for Weights.
	 */
	public static class WeightModel extends UIMASemanticProbabilityModel<Weight> {

		public WeightModel() {
		}

		@Override
		public double calculateSemanticProbability(Weight candidate) {
			// every weights below 1 g and above 1 t are extremely unlikely
			final double siValue = candidate.getSIValue();
			if (siValue != -1) {
				if (siValue < 0.001) {
					return 0.1;
				} else if (siValue > 1000) {
					return 0.1;
				}
			}
			// but everything else is equally probable.
			return 1;
		}
	}

	/**
	 * The LikelihoodModel for numbers.
	 */
	public static class NumberModel
			extends UIMASemanticProbabilityModel<de.unibi.techfak.isy.lppba.descriptors.Integer> {

		public NumberModel() {
		}
		private Weight currentWeight;
		private Age currentAge;

		protected void setCurrentWeight(Weight currentWeight) {
			this.currentWeight = currentWeight;
		}

		protected void setCurrentAge(Age currentAge) {
			this.currentAge = currentAge;
		}

		@Override
		public double calculateSemanticProbability(
				de.unibi.techfak.isy.lppba.descriptors.Integer candidate) {

			// if we have a weight, check if that candidate belongs to the weight
			// annotation.
			// If so return 0.
			if (currentWeight != null) {
				final Quantity weightValue = currentWeight.getValue();
				if (weightValue == candidate) {
					return 0;
				}
				if (weightValue instanceof Range) {
					final Range weightRange = (Range) weightValue;
					if (weightRange.getLowVal() == candidate ||
							 weightRange.getHighVal() == candidate) {
						return 0;
					}
				}
			}
			// if we have an age, check if that candidate belongs to the weight
			// annotation.
			// If so return 0.
			if (currentAge != null) {
				if (currentAge instanceof ExactAge) {
					final Quantity ageValue = ((ExactAge) currentAge).getTime()
							.getValue();
					if (ageValue == candidate) {
						return 0;
					}
					if (ageValue instanceof Range) {
						final Range ageRange = (Range) ageValue;
						if (ageRange.getLowVal() == candidate ||
								 ageRange.getHighVal() == candidate) {
							return 0;
						}
					}
				}
			}

			final int value = candidate.getValue();
			// we regard the probability of a negative group size as 0.
			if (value < 0) {
				return 0;
			}
			// integers should be more probable that are in the range of 10-100.
			if (value >= 10 && value <= 100) {
				return 0.5;
			}
			// Integers generally are probably belonging to another annotation.
			return 0.2;
		}
	}

	public static class LabGroupCompositor extends UIMACompositor<LabGroup> {

		private SyntacticProbabilityModel syntacticModel;

		public LabGroupCompositor() {
		}

		public void setSyntacticModel(SyntacticProbabilityModel syntacticModel) {
			this.syntacticModel = syntacticModel;
		}

		public SyntacticProbabilityModel getSyntacticModel() {
			return syntacticModel;
		}

		@Override
		public LabGroup UIMACompose(Annotation core,
				ArrayList<ProbableHypothesis> slotCandidates, int minBegin, int maxEnd,
				JCas jcas) {
			final ProbableHypothesis<UIMAAnnotationCandidate> ageHypo = slotCandidates
					.get(0);
			final ProbableHypothesis<UIMAAnnotationCandidate> weightHypo = slotCandidates
					.get(2);
			// find the most probable number hypothesis.
			if (ageHypo != null && ageHypo.getCandidate() != null) {
				NUMBERMODEL.setCurrentAge((Age) ageHypo.getCandidate().getAnnotation());
			}
			if (weightHypo != null && weightHypo.getCandidate() != null) {
				NUMBERMODEL.setCurrentWeight((Weight) weightHypo.getCandidate()
						.getAnnotation());
			}
			final ProbableHypothesis<UIMAAnnotationCandidate<Integer>> maxProbNumber =
					UIMAAnnotationCompositionUtil
					.calculateMaxProbHypothesis(core, syntacticModel, NUMBERMODEL,
					de.unibi.techfak.isy.lppba.descriptors.Integer.class, jcas);
			// update maxEnd and minBegin
			if (maxProbNumber != null && maxProbNumber.getCandidate() != null) {
				minBegin = Math.min(minBegin, maxProbNumber.getCandidate().getBegin());
				maxEnd = Math.max(maxEnd, maxProbNumber.getCandidate().getEnd());
			}
			// calculate the overall probability for this annotation which we
			// arbitrarily define as
			// the average probability of the annotations divided by a normalization
			// factor.
			double probability = 0;
			for (final ProbableHypothesis<UIMAAnnotationCandidate> slotHypo : slotCandidates) {
				if (slotHypo != null) {
					probability += slotHypo.getProbability();
				}
			}
			if (maxProbNumber != null) {
				probability += maxProbNumber.getProbability();
			}
			probability = probability / 4;
			if (core instanceof SpecialLaboratoryAnimal) {
				// we give a probability bonus for a Laboratory Rat annotation
				probability += 0.05;
			}
			// and we cut off the probability at 1.
			if (probability > 1) {
				probability = 1;
			}
			// if the composition probability is bigger than our threshold, store
			// this result
			if (probability > THRESHOLD) {
				// create annotation
				final LabGroup lg = new LabGroup(jcas, minBegin, maxEnd);
				lg.setOrganism((Organism) core);
				// get all slot fillers

				if (ageHypo != null && ageHypo.getCandidate() != null) {
					lg.setAge((Age) ageHypo.getCandidate().getAnnotation());
				}
				final ProbableHypothesis<UIMAAnnotationCandidate> genderHypo = slotCandidates
						.get(1);
				if (genderHypo != null && genderHypo.getCandidate() != null) {
					lg.setGender((Gender) genderHypo.getCandidate().getAnnotation());
				}
				if (weightHypo != null && weightHypo.getCandidate() != null) {
					lg.setWeight((Weight) weightHypo.getCandidate().getAnnotation());
				}
				if (maxProbNumber != null && maxProbNumber.getCandidate() != null) {
					lg.setNumber(maxProbNumber.getCandidate().getAnnotation());
				}
				lg.setProbability(probability);
				lg.setAnnotationId(AnnotationCounter.getUniqueId());
				lg.addToIndexes();
				return lg;
			} else {
				return null;
			}
		}
	}
}
