/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.injury;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import de.unibi.techfak.isy.lppba.descriptors.Quantity;
import de.unibi.techfak.isy.lppba.descriptors.Range;
import de.unibi.techfak.isy.lppba.descriptors.VertebralPosition;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import de.unibi.techfak.scie.util.CachedJCasUtil;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.jcas.JCas;

/**
 * This annotates a Vertebtral position in T-notation (e.g. "T9" for the ninth disc of the vertebrae
 * column). It uses Integer-Annotations done by the NumberAnnotator.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class VertebralPositionAnnotator extends JCasAnnotator_ImplBase {

	public static final int CHECKRANGE = 7;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);

		final Collection<Quantity> quantities = jcasUtil.select(Quantity.class);
		final String docText = jcas.getDocumentText();
		//we only consider ranges and integer ranges.
		final HashSet<Quantity> relevantQuantities = new HashSet<>();
		for (final Quantity q : quantities) {
			if (q instanceof de.unibi.techfak.isy.lppba.descriptors.Integer) {
				/* 
				 * we make a pre-check here: If the value is below 1 or greater than 12, it can't be
				 * part of a valid descriptor.
				 */
				final int val = ((de.unibi.techfak.isy.lppba.descriptors.Integer) q).getValue();
				if (val < 1 || val > 12) {
					continue;
				}
				relevantQuantities.add(q);
			} else if (q instanceof Range &&
					((Range) q).getLowVal() instanceof de.unibi.techfak.isy.lppba.descriptors.Integer &&
					((Range) q).getHighVal() instanceof de.unibi.techfak.isy.lppba.descriptors.Integer) {
				final int lowVal =
						((de.unibi.techfak.isy.lppba.descriptors.Integer) ((Range) q).
						getLowVal()).getValue();
				final int highVal =
						((de.unibi.techfak.isy.lppba.descriptors.Integer) ((Range) q).
						getHighVal()).getValue();
				if (lowVal < 1 || lowVal >= highVal || highVal > 12) {
					continue;
				}
				relevantQuantities.add(q);
			}
		}

		for (final Quantity q : relevantQuantities) {
			//we only consider quantities that are not covered by another one.
			final List<Quantity> coverings =
					jcasUtil.selectCovering(Quantity.class, q.getBegin(), q.getEnd());
			boolean existsCovering = false;
			for (final Quantity covering : coverings) {
				if (covering != q && relevantQuantities.contains(covering)) {
					existsCovering = true;
					break;
				}
			}
			if (existsCovering) {
				continue;
			}

			/* 
			 * we retrieve the high value of a range/the value of the integer here again to ensure
			 * that the descriptor stays valid, even if we have the additional information about
			 * the spinal chord region.
			 */
			final int highVal;
			if (q instanceof de.unibi.techfak.isy.lppba.descriptors.Integer) {
				highVal = ((de.unibi.techfak.isy.lppba.descriptors.Integer) q).
						getValue();
			} else {
				highVal =
						((de.unibi.techfak.isy.lppba.descriptors.Integer) ((Range) q).
						getHighVal()).getValue();
			}
			/*
			 * We search for the previous token and look if it is a valid descriptor.
			 */
			if (q.getBegin() == 0) {
				continue;
			}
			final List<Token> previousTokens =
					jcasUtil.selectCovered(Token.class, Math.max(q.
					getBegin() - CHECKRANGE, 0), q.getBegin());
			previousTokens.addAll(jcasUtil.selectCovering(Token.class, q.getBegin(), q.getEnd()));
			if (previousTokens.isEmpty()) {
				continue;
			}
			Collections.reverse(previousTokens);
			Iterator<Token> iterator = previousTokens.iterator();
			/*
			 * Now iterate over the reversed list of found tokens and try to find the first token that
			 * starts before the quantity.
			 */
			Token relevantToken = null;
			while (iterator.hasNext()) {
				final Token token = iterator.next();
				if (token.getBegin() < q.getBegin()) {
					relevantToken = token;
					break;
				}
			}
			if (relevantToken == null) {
				continue;
			}
			/*
			 * Now we have two cases: Either the token overlaps with our quantity (then we have to
			 * consider the string before our quantity or it does not and we consider the whole token).
			 */
			String prev = docText.substring(relevantToken.getBegin(),
					Math.min(relevantToken.getEnd(), q.getBegin()));
			int shift=0;
			if(prev.contains("-")&&!prev.equals("-")){
				String[] splittet=prev.split("-");
				shift=prev.length();
				prev=splittet[splittet.length-1];
				shift=shift-prev.length();
			}
			final String region;
			switch (prev) {
				case "T":
				case "Th":
					region = "Thoracic";
					break;
				case "C":
					if (highVal > 7) {
						continue;
					}
					region = "Cervical";
					break;
				case "L":
					if (highVal > 5) {
						continue;
					}
					region = "Lumbar";
					break;
				default:
					region = null;
			}
			if (region != null) {
				final VertebralPosition vp = new VertebralPosition(jcas,
						relevantToken.getBegin()+shift, q.getEnd());
				vp.setProbability(1);
				vp.setNumber(q);
				vp.setRegion(region);
				vp.setAnnotationId(AnnotationCounter.getUniqueId());
				vp.addToIndexes();
			}
		}
	}
}
