/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.quantities;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.jcas.JCas;

/**
 * This annotates Integers and Floating point numbers in the text and creates
 * Integer and Double annotations out of it. The matching is based on regular
 * expressions and does not depend on other Annotators.
 * 
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class NumberAnnotator extends
		org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public NumberAnnotator() {
	}

	public static final Pattern PLAININT = Pattern.compile("\\-?\\d+");
	public static final Pattern DOTINT = Pattern
			.compile("\\-?\\d{1,3}[\\.,](\\d{3}[\\.,])*\\d{3}");
	public static final Pattern ABBRINT = Pattern
			.compile("(\\-?\\d+)\\s*([KMkm])[.!?,;\\s]");
	public static final Pattern[] INTPATTERNS = { PLAININT, DOTINT, ABBRINT };
	public static final Pattern PLAINDOUBLE = Pattern
			.compile("\\-?\\d+[.,]\\s*\\d+");
	public static final Pattern DOUBLEE = Pattern
			.compile("(\\-?\\d+(?:[.,]\\s*\\d+)?)\\s*E(\\-?\\d+)");
	public static final Pattern ABBRDOUBLE = Pattern
			.compile("(\\-?\\d+[.,]\\s*\\d+)\\s*([KMkm])[.!?,;\\s]");
	public static final Pattern DOUBLEEXP = Pattern
			.compile("(\\-?\\d+(?:[.,]\\s*\\d+)?)\\s*\\*\\s*10\\s*\\^?\\s*(\\-?\\d+)");
	public static final Pattern[] DOUBLEPATTERNS = { PLAINDOUBLE, ABBRDOUBLE,
			DOUBLEE, DOUBLEEXP };

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final String input = jcas.getDocumentText();
		for (int p = 0; p < INTPATTERNS.length; p++) {
			final Matcher matcher = INTPATTERNS[p].matcher(input);
			while (matcher.find()) {
				int val = 0;
				int start = matcher.start();
				int end = matcher.end();
				try {
					switch (p) {
					case 0:
						val = Integer.parseInt(matcher.group());
						break;
					case 1:
						val = Integer.parseInt(matcher.group().replaceAll("[\\.,]", ""));
						break;
					case 2:
						end = matcher.end(2);
						final String numStr = matcher.group(1);
						final String magnStr = matcher.group(2);
						switch (magnStr.charAt(0)) {
						case 'M':
						case 'm':
							val = Integer.parseInt(numStr) * 1000000;
							break;
						case 'K':
						case 'k':
							val = Integer.parseInt(numStr) * 1000;
							break;
						}
						break;
					}
				} catch (NumberFormatException ex) {
					// too big numbers (> 2^31 -1) have no value
					System.err.println("Warning! " + matcher.group()
							+ " has no value because the number is too large!");
				} finally {
					// ... but they should be tagged anyhow.
					final de.unibi.techfak.isy.lppba.descriptors.Integer i = new de.unibi.techfak.isy.lppba.descriptors.Integer(
							jcas);
					i.setValue(val);
					i.setBegin(start);
					i.setEnd(end);
					i.setAnnotationId(AnnotationCounter.getUniqueId());
					i.addToIndexes();
					// if the value < 0 tag the unsigned version also
					if (val < 0) {
						final de.unibi.techfak.isy.lppba.descriptors.Integer unsignedI = new de.unibi.techfak.isy.lppba.descriptors.Integer(
								jcas);
						unsignedI.setValue(-val);
						unsignedI.setBegin(start + 1);
						unsignedI.setEnd(end);
						unsignedI.setAnnotationId(AnnotationCounter.getUniqueId());
						unsignedI.addToIndexes();
					}
				}
			}
		}
		for (int p = 0; p < DOUBLEPATTERNS.length; p++) {
			final Matcher matcher = DOUBLEPATTERNS[p].matcher(input);
			while (matcher.find()) {
				double val = 0;
				int start = matcher.start();
				int end = matcher.end();
				switch (p) {
				case 0:
					val = Double.parseDouble(preprocessRawDoubleString(matcher.group()));
					break;
				case 1:
					end = matcher.end(2);
					final String numStr = preprocessRawDoubleString(matcher.group(1));
					final String magnStr = matcher.group(2);
					switch (magnStr.charAt(0)) {
					case 'M':
					case 'm':
						val = Double.parseDouble(numStr) * 1000000;
						break;
					case 'K':
					case 'k':
						val = Double.parseDouble(numStr) * 1000;
						break;
					}
					break;
				case 2:
				case 3:
					final double factor = Double
							.parseDouble(preprocessRawDoubleString(matcher.group(1)));
					final double exp = Double.parseDouble(matcher.group(2));
					val = factor * Math.pow(10, exp);
					break;
				}
				final de.unibi.techfak.isy.lppba.descriptors.Double d = new de.unibi.techfak.isy.lppba.descriptors.Double(
						jcas);
				d.setValue(val);
				d.setBegin(start);
				d.setEnd(end);
				d.setAnnotationId(AnnotationCounter.getUniqueId());
				d.addToIndexes();
				// if the value < 0 tag the unsigned version also
				if (val < 0) {
					final de.unibi.techfak.isy.lppba.descriptors.Double unsignedD = new de.unibi.techfak.isy.lppba.descriptors.Double(
							jcas);
					unsignedD.setValue(-val);
					unsignedD.setBegin(start + 1);
					unsignedD.setEnd(end);
					unsignedD.setAnnotationId(AnnotationCounter.getUniqueId());
					unsignedD.addToIndexes();
				}
			}
		}
	}

	public String preprocessRawDoubleString(final String matchedString) {
		return matchedString.replaceAll("\\s+", "").replaceAll(",", ".");
	}
}
