/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.quantities;

import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

import de.unibi.techfak.isy.lppba.descriptors.Quantity;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import de.unibi.techfak.scie.util.CachedJCasUtil;

/**
 * This annotates quantities that have a unit following after them. This
 * requires that the NumberAnnotator and/or the WordNumberAnnotator have been
 * used already. It is an abstract parent class for any annotator that requires
 * a unit, e.g. Time, Speed, Length, Dose ...
 * 
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public abstract class UnitQuantityAnnotator extends
		org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	private final Unit[] units;
	private final int followingTokens;

	/**
	 * UnitQuantityAnnotator that compares the &#060;followingTokens&#062;
	 * following Tokens after each Quantity. The comparison is made with the
	 * matcher from &#060;units&#062;
	 * 
	 * @param units
	 *          , the units to match with this Annotator
	 * @param followingTokens
	 *          , the Number of tokens that are compared to the unit
	 */
	public UnitQuantityAnnotator(Unit[] units, int followingTokens) {
		this.units = units;
		this.followingTokens = followingTokens;
	}

	/**
	 * UnitQuantityAnnotator that compares the first Token after each Quantity.
	 * The comparison is made with the matcher from &#060;units&#062;
	 * 
	 * @param units
	 *          , the units to match with this Annotator
	 */
	public UnitQuantityAnnotator(Unit[] units) {
		this.units = units;
		this.followingTokens = 1;
	}

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);

		final Collection<Quantity> quantities = jcasUtil.select(Quantity.class);
		for (final Quantity q : quantities) {
			// look if there is a covering quantity. If so we want that one.
			final List<Quantity> coverings = jcasUtil.selectCovering(Quantity.class,
					q.getBegin(), q.getEnd());
			boolean existsCovering = false;
			for (final Quantity covering : coverings) {
				if (covering != q) {
					existsCovering = true;
					break;
				}
			}
			if (existsCovering) {
				continue;
			}
			//look for this token(to get e.g. "3mg")
			
			final List<Token> current=jcasUtil.selectOverlapping(Token.class, q.getBegin(), q.getEnd());
			StringBuilder buffer = new StringBuilder();
			int realFollowingTokens=followingTokens;
			Token cur=null;
			boolean isFinished=true;
			if((!current.isEmpty())){
				cur=current.get(current.size()-1);
				if(q.getEnd()!=cur.getEnd()){
					buffer.append(jcas.getDocumentText().substring(q.getEnd(), cur.getEnd()).toLowerCase());
					realFollowingTokens--;
					isFinished=false;
				}
			}
			
			// look for the next token
			final List<Token> following = JCasUtil.selectFollowing(jcas, Token.class,
					q, realFollowingTokens);
			if(!isFinished){
				following.add(0,cur);
			}
			if (following != null && !following.isEmpty()) {
				for (int i=followingTokens-realFollowingTokens;i<following.size();i++) {
					Token token=following.get(i);
					buffer.append(token.getCoveredText().toLowerCase());
				}
				final String unitText = buffer.toString();
				for (final Unit unit : units) {
					for (final Pattern p : unit.getPatterns()) {
						final Matcher matcher = p.matcher(unitText);
						if (matcher.matches()) {
							handleQuantity(q, unit, following, jcas);
						}
					}
				}
			}
		}
	}

	public abstract void handleQuantity(final Quantity q, final Unit unit,
			List<Token> unitTokens, JCas jcas);
}
