/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.result;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import de.unibi.techfak.isy.lppba.descriptors.PValue;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import de.unibi.techfak.scie.util.CachedJCasUtil;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This annotates P- or significance-values like P < 0.05. It basically iterates over all doubles
 * found in the text, looks up whether they are in a valid probability range (>=0 and <=1) and then
 * looks for a "P" in front of them. If those criteria are matched the Annotator makes a
 * PValue-Annotation.
 *
 * This annotator depends on the NumberAnnotator and the JULIE Lab Token/Sentence Tokenizer.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class PValueAnnotator extends
	org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public PValueAnnotator() {
	}

	public static final int CHECKRANGE = 10;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		// Go through all doubles that were found.
		final Collection<de.unibi.techfak.isy.lppba.descriptors.Double> doubles = JCasUtil
			.select(jcas, de.unibi.techfak.isy.lppba.descriptors.Double.class);
		for (final de.unibi.techfak.isy.lppba.descriptors.Double d : doubles) {
			// this might be a p value if the double is in the range 0 - 1
			if (d.getValue() < 0 || d.getValue() > 1) {
				continue;
			}
			/*
			 * Now we check if we find the 'P' for a P-Value.
			 */
			Token pToken = null;
			//first we check if we have a construction like "P0.01", where the P is part of the same token as the double.
			final List<Token> coverings = jcasUtil.selectCovering(Token.class, d.getBegin(), d.
				getBegin() + 1);
			if (!coverings.isEmpty()) {
				final Token overlapping = coverings.get(0);
				if (overlapping.getBegin() < d.getBegin()) {
					if (overlapping.getBegin() == d.getBegin() - 1 && overlapping.getCoveredText().
						toUpperCase().startsWith("P")) {
						pToken = overlapping;
					} else {
						continue;
					}
				}
			}
			if (pToken == null) {
				/*
				 * If we did not find the 'P' token by now we look in the tokens preceding the
				 * double by at most CHECKRANGE characters.
				 */
				final List<Token> previousTokens = jcasUtil.selectCovered(Token.class,
					Math.max(d.getBegin() - CHECKRANGE, 0), d.getBegin());
				if (previousTokens.isEmpty()) {
					continue;
				}
				Collections.reverse(previousTokens);
				/*
				 * Now iterate over the reversed list of found tokens and look for a single P.
				 */
				for (final Token prev : previousTokens) {
					if (prev.getCoveredText().toUpperCase().equals("P")) {
						pToken = prev;
						break;
					}
				}
			}
			if (pToken != null) {
				final PValue pval = new PValue(jcas, pToken.getBegin(), d.getEnd());
				pval.setValue(d);
				pval.setAnnotationId(AnnotationCounter.getUniqueId());
				pval.addToIndexes();
			}
		}
	}
}
