/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.structure;

import java.util.Collection;
import java.util.List;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

import de.unibi.techfak.isy.lppba.descriptors.Header;
import de.unibi.techfak.isy.lppba.descriptors.HeaderKeyword;
import de.unibi.techfak.isy.lppba.descriptors.TextBlock;
import de.unibi.techfak.scie.util.CachedJCasUtil;
import de.unibi.techfak.scie.util.Histogramm;

/**
 * This annotates headers/Überschriften of sections in the paper text. It uses
 * the TextBlocks set by the PDFStructuredTextExtractor. A header is recognized
 * as a TextBlock that
 * <ul>
 * <li>Has a higher relative font size than 1 and</li>
 * <li>is short</li>
 * <li>contains some keywords like "materials", "method", "introduction", and so
 * on.</li>
 * </ul>
 * 
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class HeaderAnnotator extends
		org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	/**
	 * This determines how the block length is weighted relative to the font size.
	 */
	public static final double LENGTHWEIGHT = -0.02;
	/**
	 * This determines how the occurence of a HeaderKeyword is weighted relative
	 * to the font size.
	 */
	public static final double KEYWORDWEIGHT = 2;
	/**
	 * We think that a "normal" header should have less characters than this
	 * number. If it has less we regard it as more probable, with more we regard
	 * it as less probable.
	 */
	public static final int SWITCHLENGTH = 20;
	/**
	 * We regard everything as a header that has a higher confidence than
	 * THRESHOLD according to our logistic regression (see below).
	 */
	public static final double THRESHOLD = 0.5;
	/**
	 * A constant factor for the scalar product inside the logistic regression.
	 */
	public static final double SHARPNESS = 2;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final Collection<TextBlock> blocks = JCasUtil.select(jcas, TextBlock.class);
		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		for (final TextBlock block : blocks) {
			// get the block length
			final int blockLength = block.getEnd() - block.getBegin();
			final double weightedLength = (blockLength - SWITCHLENGTH) * LENGTHWEIGHT;
			// get the relative font size.
			final double relFontSize = block.getRelativeFontSize() - 1;
			// get the header keywords.
			final List<HeaderKeyword> keywords = jcasUtil.selectCovered(
					HeaderKeyword.class, block.getBegin(), block.getEnd());
			// get the normalized name of the most occuring keyword.
			final Histogramm<String> keywordHisto = new Histogramm<>();
			double accumulatedProbability = 0;
			for (final HeaderKeyword keyword : keywords) {
				if (!keyword.getNodeName().equals("Materials")) {
					keywordHisto.addDataPoint(keyword.getNodeName());
				} else {
					keywordHisto.addDataPoint("Method");
				}
				accumulatedProbability += keyword.getProbability();
			}
			final String normalizedTitle = keywordHisto.getMaxElement();
			final double keyWordBias = accumulatedProbability * KEYWORDWEIGHT;

			final double confidence = 1 / (1 + Math.exp(SHARPNESS
					* (-weightedLength - relFontSize - keyWordBias)));
			if (confidence > THRESHOLD) {
				final Header header = new Header(jcas, block.getBegin(), block.getEnd());
				header.setConfidence(confidence);
				header.setNormalizedTitle(normalizedTitle);
				header.addToIndexes();
			}
		}
	}
}