/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * The OntologyDatabase class builds the main API of the ontology database.
 * Developers soley have to instanciate an implementation of the
 * OntologyDatabaseDriver class and the OntologyDatabase class in order to
 * obtain a fully working ontology database.
 *
 * The ontology database consists of ontologies. Ontologies are represented by
 * the OntologyHandle object, such an handle can be obtained by simply calling
 * the "getOntology" function. Each ontology consists of a set of nodes. Nodes
 * can be created by instanciating the Node class and calling the "storeNode"
 * function. The node class can then be used to model associations between nodes
 * and the so called NodeNames that are associated to them. The words in the
 * "NodeNames" are automatically indexed in a dictionary that allows to retrieve
 * all nodes which have a NodeName that contains a word simmilar to a given
 * word.
 *
 * @author Andreas Stöckel -- astoecke@techfak.uni-bielefeld.de
 */
public final class OntologyDatabase implements AutoCloseable {

	/**
	 * Regular expression for valid ontology names.
	 */
	private static final String ONTOLOGY_NAME_REGEX = "[A-Za-z][A-Za-z0-9_]*";

	/**
	 * Pattern representing valid ontology names.
	 */
	private static final Pattern ONTOLOGY_NAME = Pattern.compile(ONTOLOGY_NAME_REGEX);

	/**
	 * Specifies whether the ontology database has been closed.
	 */
	private boolean closed = false;

	/**
	 * A reference to the actual database implementation.
	 */
	private OntologyDatabaseDriver driver;

	/**
	 * A reference to the NodeFactory instance. The node factory is responsible
	 * for translating NodeHandle or NodeDescriptor instances to the much more
	 * user friendly Node instances, and querying the needed NodeDescriptors
	 * from the database if they are not available. The nodeFactory never
	 * instanciates more than one instance for one Node at a time. The Node
	 * instances themselves have a reference to the nodeFactory in order to be
	 * capable of performing database queries and translating NodeHandles to
	 * Nodes.
	 */
	private NodeFactory nodeFactory;

	/**
	 * A reference to the Dictionary instance. The Dictionary contains code for
	 * updating an querying the dictionary and controlling the database
	 * (Business Logic).
	 */
	private Dictionary dictionary;

	/**
	 * Constructor of the OntologyDatabase class. An instance of a
	 * OntologyDatabaseDriver implementation is needed for using the
	 * OntologyDatabase.
	 *
	 * @param driver is a reference to the OntologyDatabaseDriver instance that
	 * should be used as a storage backend.
	 */
	public OntologyDatabase(OntologyDatabaseDriver driver) {
		this.driver = driver;
		this.nodeFactory = new NodeFactory(driver);
		this.dictionary = new Dictionary(driver);
	}

	/**
	 * Checks whether the database is currently closed.
	 *
	 * @throws RuntimeException 
	 */
	private void checkClosed() throws RuntimeException {
		if (closed) {
			throw new RuntimeException("Invalid operation, the ontology "
					+ "database is closed.");
		}
	}

	/**
	 * Internally used function that returns a reference to the node factory.
	 *
	 * @return a reference to the internally used node factory.
	 */
	NodeFactory getNodeFactory() {
		return nodeFactory;
	}

	/**
	 * Internally used function that returns a reference to the database driver.
	 * The database driver is not meant to be used by the end-user of this
	 * library. Note that the result might not be a reference to the driver
	 * provided in the constructor, the OntologyDatabase might include
	 * additional caching layers.
	 *
	 * @return a reference to the storage backend.
	 */
	OntologyDatabaseDriver getDriver() {
		return driver;
	}

	/**
	 * Returns a list containing handles to all currently available ontologies.
	 *
	 * @return a list containing references to all currently available
	 * ontologies or an empty list if no ontologies are currently available.
	 */
	public Collection<OntologyHandle> getOntologies() {
		checkClosed();
		return driver.getOntologies();
	}

	/**
	 * Returns an ontology handle for an ontology with the given name, if such
	 * an ontology does not exist yet and the "create" flag is set to true, it
	 * is created and the new ontology handle is returned.
	 *
	 * @param name is the name of the queried ontology. Note that the name has
	 * to be of the following pattern: [A-Za-z][A-Za-z0-9_]
	 * @param create specifies whether the ontology should be created if it does
	 * not exist.
	 * @return an handle to the ontology with the given name.
	 * @throws InvalidNameException if the name is not of the pattern listed
	 * above.
	 */
	public OntologyHandle getOntology(String name, boolean create) throws InvalidNameException {
		checkClosed();
		if (!ONTOLOGY_NAME.matcher(name).matches()) {
			throw new InvalidNameException("Invalid ontology name \"" + name
					+ "\", must adhere to pattern " + ONTOLOGY_NAME_REGEX);
		}
		return driver.getOntology(name, create);
	}

	/**
	 * Returns an ontology handle for an ontology with the given name, if such
	 * an ontology does not exist yet, it is created and the new ontology handle
	 * is returned.
	 *
	 * @param name is the name of the queried ontology. Note that the name has
	 * to be of the following pattern: [A-Za-z][A-Za-z0-9_]
	 * @return an handle to the ontology with the given name.
	 * @throws InvalidNameException if the name is not of the pattern listed
	 * above.
	 */
	public OntologyHandle getOntology(String name) throws InvalidNameException {
		return getOntology(name, true);
	}

	/**
	 * Returns an ontology handle for an ontology with the given name, if such
	 * an ontology does not exist yet, it is created and the new ontology handle
	 * is returned. It is ensured that the ontology is empty.
	 *
	 * @param name is the name of the queried ontology. Note that the name has
	 * to be of the following pattern: [A-Za-z][A-Za-z0-9_]
	 *
	 * @return an handle to the ontology with the given name.
	 * @throws InvalidNameException if the name is not of the pattern listed
	 * above.
	 */
	public OntologyHandle getEmptyOntology(String name) throws InvalidNameException {
		checkClosed();
		if (!ONTOLOGY_NAME.matcher(name).matches()) {
			throw new InvalidNameException("Invalid ontology name \"" + name
					+ "\", must adhere to pattern " + ONTOLOGY_NAME_REGEX);
		}
		return driver.getEmptyOntology(name);
	}

	/**
	 * Returns true if the given ontology exists.
	 *
	 * @param name is the name of the queried ontology. Note that the name has
	 * to be of the following pattern: [A-Za-z][A-Za-z0-9_]
	 * @return true if the ontology exists, false otherwise.
	 * @throws InvalidNameException if the name is not of the pattern listed
	 * above.
	 */
	public boolean queryOntologyExists(String name) throws InvalidNameException {
		checkClosed();
		if (!ONTOLOGY_NAME.matcher(name).matches()) {
			throw new InvalidNameException("Invalid ontology name \"" + name
					+ "\", must adhere to pattern " + ONTOLOGY_NAME_REGEX);
		}
		return driver.getOntology(name, false) != null;
	}

	/**
	 * Deletes the ontology with the given handle. Deletes all corresponding
	 * nodes and dictionary entries.
	 *
	 * @param ontology is the handle of the ontology that should be deleted.
	 */
	public void deleteOntology(OntologyHandle ontology) {
		checkClosed();
		driver.deleteOntology(ontology);
		nodeFactory.purgeOntology(ontology);
	}

	/**
	 * Returns the modification date of the given ontology.
	 *
	 * @param ontology is the ontology handle.
	 * @return the date of the last modification.
	 */
	public Date getOntologyModificationDate(OntologyHandle ontology) {
		checkClosed();
		return driver.getOntologyModificationDate(ontology);
	}

	/**
	 * Updates the modification date of the given ontology to now.
	 *
	 * @param ontology is the ontology handle.
	 */
	public void updateOntologyModificationDate(OntologyHandle ontology) {
		checkClosed();
		driver.setOntologyModificationDate(ontology, new Date());
	}

	/**
	 * Returns all the nodes that were marked as "root node" in the given
	 * ontology.
	 *
	 * @param ontology is the ontology for which the root nodes should be
	 * returned.
	 * @return a collection of all nodes that were marked as root nodes or null
	 * if the ontology does not exist.
	 */
	public Collection<Node> getRootNodes(OntologyHandle ontology) {
		checkClosed();
		return nodeFactory.descriptorsToNodeList(driver.getRootNodes(ontology));
	}

	/**
	 * Returns a "Node" instance for the given node handle. A node handle
	 * consists of an ontology and a node id.
	 *
	 * @param node is an arbitrary node handle.
	 * @return a node corresponding to the handle or "null" if the node does not
	 * exist.
	 */
	public Node getNode(NodeHandle node) {
		checkClosed();
		return nodeFactory.toNode(node);
	}

	/**
	 * Returns a "Node" instance for the given node handle.
	 *
	 * @param node is a high-level node instance.
	 * @return a node corresponding to the handle, never returns null, the node
	 * is not queried from the database!
	 */
	public Node getNode(NodeDescriptor node) {
		checkClosed();
		return nodeFactory.toNode(node);
	}

	/**
	 * Converts a list of NodeHandle instances to high-level "Node" instances.
	 * See getNode for the exact behaviour.
	 *
	 * @param nodes a list of NodeHandles.
	 * @return a list of "Node" instances. If a node does not exist, "null" is
	 * returned in that slot.
	 */
	public List<Node> handlesToNodeList(Collection<NodeHandle> nodes) {
		checkClosed();
		return nodeFactory.handlesToNodeList(nodes);
	}

	/**
	 * Converts a list of NodeDesciptor instances to high-level "Node"
	 * instances. See "getNode" for the exact behaviour.
	 *
	 * @param nodes a list of NodeHandles.
	 * @return a list of "Node" instances. The nodes are not queried from the
	 * database, only the data type is converted!
	 */
	public List<Node> descriptorsToNodeList(Collection<NodeDescriptor> nodes) {
		checkClosed();
		return nodeFactory.descriptorsToNodeList(nodes);
	}

	/**
	 * Stores the given NodeDescriptor in the database. If the node the
	 * descriptor refers to does not exist, it is created (as long as the
	 * refered ontology exists). Otherwise the node is only updated.
	 *
	 * @param node is the NodeDescriptor instance that should be stored in the
	 * database.
	 */
	public void storeNode(NodeDescriptor node) {
		checkClosed();
		nodeFactory.registerNode(node);
		driver.storeNode(node);
		dictionary.updateNode(node);
	}

	/**
	 * Deletes the node the given handle refers to from the database and deletes
	 * all corresponding dictionary entries.
	 *
	 * @param node is the node that should be deleted.
	 */
	public void deleteNode(NodeHandle node) {
		checkClosed();
		driver.deleteNode(node);
		nodeFactory.purgeNode(node);
	}

	/**
	 * Deletes the node the given descriptor refers to from the database and
	 * removes all corresponding dictionary entries.
	 *
	 * @param node is the node that should be deleted.
	 */
	public void deleteNode(NodeDescriptor node) {
		checkClosed();
		deleteNode(node.getHandle());
	}

	/**
	 * Queries a set of words from the database an returns a map between these
	 * words and the corresponding dictionary entries.
	 *
	 * @param words is a collection of strings (single words) that should be
	 * queried.
	 * @return a map which contains a key for each found word and the
	 * corresponding set of matches in the dictionary.
	 */
	public Map<String, Collection<DictionaryMatch>> query(Collection<String> words) {
		checkClosed();
		return dictionary.query(words);
	}

	/**
	 * Queries a single word from the dictionary.
	 *
	 * @param word is the word that should be queried.
	 * @return a set of matches or an empty set if there is no match.
	 */
	public Collection<DictionaryMatch> query(String word) {
		checkClosed();
		Collection<DictionaryMatch> res
				= query(Arrays.asList(new String[]{word})).get(word);
		if (res == null) {
			return new LinkedList();
		}
		return res;
	}

	/**
	 * Shuts down the database.
	 */
	@Override
	public void close() {
		// Close the node factory
		nodeFactory.close();
		driver.close();

		// Invalidate all references
		nodeFactory = null;
		dictionary = null;
		driver = null;

		// Mark the instance as closed
		closed = true;
	}

}
