/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

/**
 * The AssociationType enum is used to specify the type of an association.
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public enum AssociationType {

	/**
	 * The DEFAULT association type specifies no special association type --
	 * this association type may be used if the type of the association does not
	 * matter and a generic association should be expressed.
	 */
	DEFAULT(0x01),
	/**
	 * The IS_A association type specifies the source node is an instance of the
	 * target node.
	 */
	IS_A(0x02),
	/**
	 * The CONTAINS_A association type specifies that the source node contains
	 * an instance of the target node.
	 */
	CONTAINS_A(0x04),
	/**
	 * The SYNONYM_OF association type specifies that the source node is
	 * equivalent to the target node but just contains other names.
	 */
	SYNONYM_OF(0x08);

	/**
	 * Each association type is given a value that can be used as a bit mask
	 * (each value is a power of two).
	 */
	private final int value;

	/**
	 * Constructor of AssociationType, copies the given value.
	 *
	 * @param value is the value given to the type. The values should be powers
	 * of two in order to be usable as bit masks.
	 */
	private AssociationType(int value) {
		this.value = value;
	}

	/**
	 * Returns the associated value.
	 *
	 * @return the associated value.
	 */
	public int getValue() {
		return value;
	}

	/**
	 * Bit mask that represents all types.
	 */
	public static final int ALL = 0xFFFF;

	/**
	 * Converts an association type value to the corresponding AssociationType
	 * instance.
	 *
	 * @param value is the numerical value of the association.
	 * @return the corresponding AssociationType enumeration item.
	 */
	public static AssociationType getType(int value) {
		switch (value) {
			case 0x02:
				return IS_A;
			case 0x04:
				return CONTAINS_A;
			case 0x08:
				return SYNONYM_OF;
			default:
				return DEFAULT;
		}
	}

	public static AssociationType getType(String type) {
		type = type.toLowerCase();
		switch (type) {
			case "default":
				return DEFAULT;
			case "is_a":
				return IS_A;
			case "contains_a":
				return CONTAINS_A;
			case "synonym_of":
				return SYNONYM_OF;
			default:
				return null;
		}
	}

}
