/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

/**
 * ComparisonMode is used to determine the mode that is used for searching the
 * node names in the dictionary.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public enum ComparisonMode {

	/**
	 * The NodeName instance is compared fuzzyly, this means that the complete
	 * logic for fuzzy comparisons is used.
	 */
	FUZZY(0),
	/**
	 * The NodeName instance is stripped from non-alphabethic characters and
	 * converted to lowercase. Other than that a strict comparison is performed,
	 * the name has to appear in order. The words of the node name are splitted
	 * at whitespace and special characters such as "-", ":", "," etc.
	 */
	SOFT(1),
	/**
	 * The NodeName instance is stripped from special characters and converted
	 * to lowercase. Other than that a strict comparison is performed. The name
	 * has to appear in order. The words of the node name are splitted at
	 * whitespace and special characters such as "-", ":", "," etc.
	 */
	CLEAN(2),
	/**
	 * The NodeName instance is converted to lowercase before comparison. The
	 * name has to appear in order. The words of the node name are splitted at
	 * whitespace only!
	 */
	LOWERCASE(3),
	/**
	 * The NodeName instance is compared strictly, an exact string match is
	 * needed. The name has to appear in order. The words of the node name are
	 * splitted at whitespace only!
	 */
	STRICT(4),
	/**
	 * This NodeName instance is excluded from the dictionary. One can not
	 * search for this NodeName.
	 */
	NONE(5);

	/**
	 * The value with which the ComparisonMode is stored in the db.
	 */
	private final int value;

	/**
	 * The enum type constructor.
	 */
	private ComparisonMode(int value) {
		this.value = value;
	}

	/**
	 * Returns the value with which the ComparisonMode is stored in the db.
	 *
	 * @return the value with which the ComparisonMode is stored in the db.
	 */
	public int getValue() {
		return value;
	}

	/**
	 * Converts an comparison type value to the corresponding ComparisonMode
	 * instance.
	 *
	 * @param value is the numerical value of the comparison type.
	 * @return the corresponding NodeNameComparisonType enumeration item.
	 */
	public static ComparisonMode getMode(int value) {
		switch (value) {
			case 0:
				return FUZZY;
			case 1:
				return SOFT;
			case 2:
				return CLEAN;
			case 3:
				return LOWERCASE;
			case 4:
				return STRICT;
			case 5:
				return NONE;
		}
		return null;
	}

	public static ComparisonMode getMode(String mode) {
		mode = mode.toLowerCase();
		switch (mode) {
			case "fuzzy":
				return FUZZY;
			case "soft":
				return SOFT;
			case "clean":
				return CLEAN;
			case "lowercase":
				return LOWERCASE;
			case "strict":
				return STRICT;
			case "none":
				return NONE;
			default:
				return null;
		}
	}

}
