/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

import org.apache.commons.lang3.StringUtils;

/**
 * The Handle class is used to uniquely identify objects in the ontology
 * database. It is the base class for both "NodeHandle" and "OntologyHandle"
 * which represent nodes and ontologies.
 *
 * @author Andreas Stöckel -- astoecke@techfak.uni-bielefeld.de
 */
public abstract class Handle implements Comparable<Handle> {

	/**
	 * Function which returns a unique name of the handle. This name can be used
	 * to uniquely identify the resource within the ontology database. This
	 * function may never return null.
	 *
	 * @return a unique name.
	 */
	public abstract String getUniqueName();

	/**
	 * Returns the unique name.
	 *
	 * @return the unique name.
	 */
	@Override
	public String toString() {
		return this.getUniqueName();
	}

	/**
	 * Two handles equal each other if their unique name is equal.
	 *
	 * @param o the object to which the handle should be compared.
	 * @return true if the two object equal each other, false otherwise.
	 */
	@Override
	public boolean equals(Object o) {
		if ((o == null) || (o.getClass() != getClass())) {
			return false;
		}

		Handle h = (Handle) o;
		return StringUtils.equals(h.getUniqueName(), getUniqueName());
	}

	/**
	 * Returns the hash code of the unique name.
	 *
	 * @return the hash code of the unique name.
	 */
	@Override
	public int hashCode() {
		return getUniqueName().hashCode();
	}

	/**
	 * Compares the unique name of the two handles.
	 *
	 * @param h the handle to which this one should be compared.
	 * @return the result of the comparison between the two unique names.
	 */
	@Override
	public int compareTo(Handle h) {
		return this.getUniqueName().compareTo(h.getUniqueName());
	}

}
