/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

/**
 * The NodeHandle class is used to uniquely identify nodes in the ontology
 * database. Each node handle consists of a reference to an ontology handle and
 * an unique ID of the node. Each ontology database has to support this scheme,
 * NodeHandle instances may also be created by the end-user of the ontology
 * database (in contrast to OntologyHandle instances, which must be created by
 * the database).
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public final class NodeHandle extends Handle {

	/**
	 * Contains a reference to the ontology handle. The NodeHandle class
	 * supports a value of null in this field, yet this support was only added
	 * to facilitate testing. In the actual application the ontology should
	 * always be non-null.
	 */
	private final OntologyHandle ontology;

	/**
	 * Contains the unique ID of the ontology node.
	 */
	private final int id;

	/**
	 * Constructor of the NodeHandle class.
	 *
	 * @param ontology is a reference to the ontology handle. The NodeHandle
	 * class supports a value of null in this field, yet this support was only
	 * added to facilitate testing. In the actual application the ontology
	 * should always be non-null.
	 * @param id is the unique ID of the ontology node.
	 */
	public NodeHandle(OntologyHandle ontology, int id) {
		this.ontology = ontology;
		this.id = id;
	}

	/**
	 * Returns the ontology handle.
	 *
	 * @return the ontology handle.
	 */
	public OntologyHandle getOntology() {
		return ontology;
	}

	/**
	 * Returns the unique id of the ontology node.
	 *
	 * @return the unique id of the ontology node.
	 */
	public int getId() {
		return id;
	}

	/**
	 * Returns a unique name of the scheme "ontology:id", or just "id" if the
	 * ontology is null (which does not happen in a real world application).
	 *
	 * @return the unique name.
	 */
	@Override
	public String getUniqueName() {
		return (ontology != null)
				? (ontology.getUniqueName() + ":" + id)
				: String.valueOf(id);
	}

}
