/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

/**
 * The OntologyHandle class is used to uniquely identify ontologies in the
 * ontology database. Ontologies have to have a unique name. Implementations of
 * DatabaseDriver have to override this class and may attach additional
 * information to the ontologies. Therefore OntologyHandles must not be created
 * by other entities than the DatabaseDriver.
 *
 * @author Andreas Stöckel -- astoecke@techfak.uni-bielefeld.de
 */
public class OntologyHandle extends Handle {

	/**
	 * Contains the name of the ontology.
	 */
	private final String name;

	/**
	 * Creates a new instance of OntologyHandle and copies the given name. This
	 * constructor has to be called from the child classes which are implemented
	 * in the DatabaseDriver implementation. The name uniquely represents the
	 * ontology.
	 *
	 * @param name is the name of the ontology. The high-level abstraction
	 * ensures that ontologies with names that do not adhere to the pattern
	 * [A-Za-z_][A-Za-z0-9_] are never created.
	 */
	protected OntologyHandle(String name) {
		this.name = name;
	}

	/**
	 * Returns the name of the ontology.
	 *
	 * @return the name of the ontology.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the name of the ontology, which uniquely represents the ontology.
	 *
	 * @return the name of the ontology.
	 */
	@Override
	public final String getUniqueName() {
		return name;
	}

}
