/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver.implementation;

import java.sql.Connection;

/**
 * Contains the default (generic) SQL statements -- this class should be
 * overwritten if database-specific SQL statements are needed.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
abstract class DefaultSQLStatements extends SQLStatements {

	protected static final String TBL_PREFIX = "ontology_";
	protected static final String TBL_ONTOLOGIES = TBL_PREFIX + "ids";
	protected static final String TBL_NODES = TBL_PREFIX + "nodes";
	protected static final String TBL_NODE_NAMES = TBL_PREFIX + "node_names";
	protected static final String TBL_ASSOCIATIONS = TBL_PREFIX + "associations";
	protected static final String TBL_DICTIONARY = TBL_PREFIX + "dictionary";

	protected static final String ALL_TABLES[] = {
		TBL_ONTOLOGIES, TBL_NODES, TBL_NODE_NAMES, TBL_ASSOCIATIONS,
		TBL_DICTIONARY
	};

	public DefaultSQLStatements(Connection connection) {
		super(connection);
	}

	@Override
	public String[] getTables() {
		return ALL_TABLES;
	}

	@Override
	public String getSQLDropTableIfExists(String table) {
		return "DROP TABLE IF EXISTS " + table;
	}

	@Override
	public String getStatementSQL(SID statementID) {
		switch (statementID) {
			case getOntologies:
				return "SELECT name, id FROM " + TBL_ONTOLOGIES;
			case getOntologyByName:
				return "SELECT id FROM " + TBL_ONTOLOGIES + " WHERE name = ?";
			case getOntologyByID:
				return "SELECT name FROM " + TBL_ONTOLOGIES + " WHERE id = ?";
			case getOntologyDate:
				return "SELECT mdate FROM " + TBL_ONTOLOGIES + " WHERE id = ?";
			case setOntologyDate:
				return "UPDATE " + TBL_ONTOLOGIES + " SET mdate = ? WHERE id = ?";
			case insertOntology:
				return "INSERT INTO " + TBL_ONTOLOGIES + " (name) VALUES (?)";
			case deleteOntology:
				return "DELETE FROM " + TBL_ONTOLOGIES + " WHERE id = ?";
			case deleteNodesForOntology:
				return "DELETE FROM " + TBL_NODES + " WHERE oid = ?";
			case deleteNodeNamesForOntology:
				return "DELETE FROM " + TBL_NODE_NAMES + " WHERE oid = ?";
			case deleteAssociationsForOntology:
				return "DELETE FROM " + TBL_ASSOCIATIONS + " WHERE oid = ?";
			case deleteDictionaryEntriesForOntology:
				return "DELETE FROM " + TBL_DICTIONARY + " WHERE oid = ?";
			case getRootNodes:
				return "SELECT id FROM " + TBL_NODES + " WHERE (oid = ?) AND (root = TRUE)";
			case getIsRootNode:
				return "SELECT root FROM " + TBL_NODES + " WHERE (oid = ?) AND (id = ?) LIMIT 1";
			case getNodeNames:
				return "SELECT name, type, mode, prim FROM " + TBL_NODE_NAMES + " WHERE (oid = ?) AND (id = ?)";
			case getNodeExists:
				return "SELECT COUNT(*) FROM " + TBL_NODES + " WHERE (oid = ?) AND (id = ?) LIMIT 1";
			case insertNodeID:
				return "INSERT INTO " + TBL_NODES + " (oid, id, root) VALUES (?, ?, ?)";
			case deleteNodeID:
				return "DELETE FROM " + TBL_NODES + " WHERE (oid = ?) AND (id = ?)";
			case insertNodeName:
				return "INSERT INTO " + TBL_NODE_NAMES + " (oid, id, name, type, mode, prim) VALUES (?, ?, ?, ?, ?, ?)";
			case deleteNodeNames:
				return "DELETE FROM " + TBL_NODE_NAMES + " WHERE (oid = ?) AND (id = ?)";
			case getInboundAssociations:
				return "SELECT src, tar, type FROM " + TBL_ASSOCIATIONS + " WHERE (oid = ?) AND (tar = ?) AND (type & ? > 0)";
			case getOutboundAssociations:
				return "SELECT src, tar, type FROM " + TBL_ASSOCIATIONS + " WHERE (oid = ?) AND (src = ?) AND (type & ? > 0)";
			case getUndirectedAssociations:
				return "SELECT src, tar, type FROM " + TBL_ASSOCIATIONS + " WHERE (oid = ?) AND ((src = ?) OR (tar = ?)) AND (type & ? > 0)";
			case insertAssociation:
				return "INSERT INTO " + TBL_ASSOCIATIONS + " (oid, src, tar, type) VALUES (?, ?, ?, ?)";
			case deleteAssociation:
				return "DELETE FROM " + TBL_ASSOCIATIONS + " WHERE (oid = ?) AND (src = ?) AND (tar = ?)";
			case deleteNodeAssociations:
				return "DELETE FROM " + TBL_ASSOCIATIONS + " WHERE (oid = ?) AND ((src = ?) OR (tar = ?))";
			case queryDictionary:
				return "SELECT oid, id FROM " + TBL_DICTIONARY + " WHERE (entry = ?)";
			case insertDictionaryEntry:
				return "INSERT INTO " + TBL_DICTIONARY + " (oid, id, entry) VALUES (?, ?, ?)";
			case deleteDictionaryEntry:
				return "DELETE FROM " + TBL_DICTIONARY + " WHERE (oid = ?) AND (id = ?)";
			default:
				throw new AssertionError(statementID.name());
		}
	}

}
