/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver.implementation;

import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.Association;
import de.unibi.techfak.scie.ontology.driver.AssociationDirection;
import de.unibi.techfak.scie.ontology.driver.AssociationType;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

class MemoryDatabaseOntologyHandle extends OntologyHandle {

	public MemoryDatabaseOntologyHandle(String name) {
		super(name);
	}

}

/**
 *
 * @author andreas
 */
public class MemoryDatabaseDriver extends OntologyDatabaseDriver {

	/**
	 * The JDBCOntologyHandle class is the OntologyHandle type used by the JDBC
	 * database. Despite having a name it also contains a mapping to the
	 * internally used id.
	 */
	protected class MemoryDatabaseOntologyHandle extends OntologyHandle {

		/**
		 * Modification date of the ontology.
		 */
		private Date modificationDate;

		/**
		 * Constructor of the JDBCOntologyHandle class.
		 *
		 * @param name is the name of the ontology.
		 */
		public MemoryDatabaseOntologyHandle(String name) {
			super(name);
			modificationDate = new Date();
		}

		public Date getModificationDate() {
			return modificationDate;
		}

		public void setModificationDate(Date modificationDate) {
			this.modificationDate = modificationDate;
		}	


	}

	private final Map<String, OntologyHandle> ontologies;

	private final Map<OntologyHandle, Map<Integer, NodeDescriptor>> nodes;

	private final Map<OntologyHandle, Set<Integer>> rootNodes;

	private final Map<OntologyHandle, Map<Integer, Set<Association>>> inboundAssociations;

	private final Map<OntologyHandle, Map<Integer, Set<Association>>> outboundAssociations;

	private final Map<String, Set<NodeHandle>> dictionary;

	private final Map<NodeHandle, Set<String>> dictionaryNodeIndex;

	public MemoryDatabaseDriver() {
		this.ontologies = new HashMap();
		this.nodes = new HashMap();
		this.rootNodes = new HashMap();
		this.inboundAssociations = new HashMap();
		this.outboundAssociations = new HashMap();
		this.dictionary = new HashMap();
		this.dictionaryNodeIndex = new HashMap();
	}

	private Collection<NodeDescriptor> nodeDescriptorsFromIds(
			OntologyHandle ontology, Collection<Integer> ids) {
		if (nodes.containsKey(ontology)) {
			Map<Integer, NodeDescriptor> idMap = nodes.get(ontology);
			Collection<NodeDescriptor> res = new ArrayList(ids.size());
			for (Integer id : ids) {
				if (idMap.containsKey(id)) {
					res.add(idMap.get(id));
				} else {
					res.add(null);
				}
			}
			return res;
		}
		return null;
	}

	private void appendAssociations(Set<Association> target,
			Map<Integer, Set<Association>> source, int id,
			int types) {
		if (source != null) {
			// Fetch all associations for the given node id
			Collection<Association> assocs = source.get(id);
			if (assocs != null) {
				// Filter the associations by their type and add them to the
				// result
				for (Association assoc : assocs) {
					if ((types & assoc.getType().getValue()) != 0) {
						target.add(assoc);
					}
				}
			}
		}
	}

	private void storeDirectedAssociation(
			Map<OntologyHandle, Map<Integer, Set<Association>>> map,
			NodeHandle node, Association assoc) {
		// Fetch the id and the corresponding ontology of the given node
		int id = node.getId();
		OntologyHandle ontology = node.getOntology();

		// Make sure the given node exists, if not, create it
		Map<Integer, NodeDescriptor> nodeMap = nodes.get(ontology);
		if (nodeMap != null && !nodeMap.containsKey(id)) {
			nodeMap.put(id, new NodeDescriptor(new NodeHandle(ontology, id),
					(Collection<NodeName>)null));
		}

		// Fetch the map for the ontology
		Map<Integer, Set<Association>> idMap = map.get(ontology);
		if (idMap != null) {
			// Fetch the set describing the associations of the id
			Set<Association> assocSet = idMap.get(id);
			if (assocSet == null) {
				assocSet = new HashSet();
				idMap.put(id, assocSet);
			}

			// Store the given association in the set
			assocSet.add(assoc);
		}
	}

	private void deleteDirectedAssociation(
			Map<OntologyHandle, Map<Integer, Set<Association>>> map,
			NodeHandle node, Association assoc) {
		// Fetch the id and the corresponding ontology of the given node
		int id = node.getId();
		OntologyHandle ontology = node.getOntology();

		// Fetch the map for the ontology
		Map<Integer, Set<Association>> idMap = map.get(ontology);
		if (idMap != null) {
			// Fetch the set describing the associations of the id
			Set<Association> assocSet = idMap.get(id);
			if (assocSet != null) {
				assocSet.remove(assoc);
			}
		}
	}

	@Override
	public Collection<OntologyHandle> getOntologies() {
		return ontologies.values();
	}

	@Override
	public OntologyHandle getOntology(String name, boolean create) {
		// Return the reference to the ontology if the ontology already exists
		if (ontologies.containsKey(name)) {
			return ontologies.get(name);
		} else if (!create) {
			return null;
		}

		// Create the ontology handle and all corresponding indexing structures
		OntologyHandle res = new MemoryDatabaseOntologyHandle(name);
		ontologies.put(name, res);
		nodes.put(res, new HashMap());
		rootNodes.put(res, new TreeSet());
		inboundAssociations.put(res, new HashMap());
		outboundAssociations.put(res, new HashMap());

		// Return the newly created ontology handle
		return res;
	}

	@Override
	public Date getOntologyModificationDate(OntologyHandle ontology) {
		return ((MemoryDatabaseOntologyHandle) ontology).getModificationDate();
	}

	@Override
	public void setOntologyModificationDate(OntologyHandle ontology, Date date) {
		((MemoryDatabaseOntologyHandle) ontology).setModificationDate(date);
	}

	@Override
	public void deleteOntology(OntologyHandle ontology) {
		// Delete all nodes which belong to this ontology from the dictionary
		if (nodes.containsKey(ontology)) {
			deleteDictionaryEntries(nodeHandlesFromDescriptors(
					nodes.get(ontology).values()));
		}

		// Try to delete all other references to the given ontology
		ontologies.remove(ontology.getName());
		nodes.remove(ontology);
		rootNodes.remove(ontology);
		inboundAssociations.remove(ontology);
		outboundAssociations.remove(ontology);
	}

	@Override
	public Collection<NodeDescriptor> getRootNodes(OntologyHandle ontology) {
		if (rootNodes.containsKey(ontology)) {
			return nodeDescriptorsFromIds(ontology, rootNodes.get(ontology));
		}
		return null;
	}

	@Override
	public List<NodeDescriptor> getNodes(Collection<NodeHandle> handles) {
		List<NodeDescriptor> res = new ArrayList(handles.size());
		for (NodeHandle handle : handles) {
			OntologyHandle ontology = handle.getOntology();
			Map<Integer, NodeDescriptor> idMap = nodes.get(ontology);
			if (idMap != null && idMap.containsKey(handle.getId())) {
				res.add(idMap.get(handle.getId()));
			} else {
				res.add(null);
			}
		}
		return res;
	}

	@Override
	public List<Boolean> queryNodesExist(Collection<NodeHandle> handles) {
		List<Boolean> res = new ArrayList(nodes.size());
		for (NodeHandle handle : handles) {
			OntologyHandle ontology = handle.getOntology();
			res.add(nodes.containsKey(ontology)
					&& (nodes.get(ontology).containsKey(handle.getId())));
		}
		return res;
	}

	@Override
	public void storeNodes(Collection<NodeDescriptor> descriptors) {
		for (NodeDescriptor descriptor : descriptors) {
			// Fetch the id and the corresponding ontology of the given
			// descriptor
			int id = descriptor.getId();
			OntologyHandle ontology = descriptor.getOntology();

			// Store the descriptor in the id map
			Map<Integer, NodeDescriptor> idMap = nodes.get(ontology);
			if (idMap != null) {
				idMap.put(descriptor.getId(), descriptor);
			}

			// Update the root node index
			Set<Integer> rootSet = rootNodes.get(ontology);
			if (rootSet != null) {
				if (descriptor.isRoot()) {
					rootSet.add(id);
				} else {
					rootSet.remove(id);
				}
			}
		}
	}

	@Override
	public void deleteNodes(Collection<NodeHandle> nodes) {
		// Delete the nodes from the dictionary
		deleteDictionaryEntries(nodes);

		// Delete the nodes from all other structures		
		for (NodeHandle node : nodes) {
			// Fetch the id and the corresponding ontology of the given
			// descriptor
			int id = node.getId();
			OntologyHandle ontology = node.getOntology();

			// Remove the node from the node index
			if (this.nodes.containsKey(ontology)) {
				this.nodes.get(ontology).remove(id);
			}

			// Remove the node from the root node index
			if (this.rootNodes.containsKey(ontology)) {
				this.rootNodes.get(ontology).remove(id);
			}

			// Remove all associations that contain this node
			final List<NodeHandle> nodeList = Arrays.asList(
					new NodeHandle[]{node});
			final List<Collection<Association>> assocs = getAssociations(
					nodeList, AssociationType.ALL,
					AssociationDirection.BOTH);
			this.deleteAssociations(assocs.get(0));
		}
	}

	@Override
	public List<Collection<Association>> getAssociations(
			Collection<NodeHandle> nodes, int associationTypes,
			AssociationDirection dir) {
		List<Collection<Association>> res = new ArrayList();
		for (NodeHandle node : nodes) {
			// Fetch the id and the corresponding ontology of the given
			// descriptor
			int id = node.getId();
			OntologyHandle ontology = node.getOntology();

			// Create the result list and fill it with both outbound and inbound
			// associations.
			Set<Association> assocs = new HashSet();
			if (dir == AssociationDirection.INBOUND
					|| dir == AssociationDirection.BOTH) {
				appendAssociations(assocs,
						inboundAssociations.get(ontology), id,
						associationTypes);
			}
			if (dir == AssociationDirection.OUTBOUND
					|| dir == AssociationDirection.BOTH) {
				appendAssociations(assocs,
						outboundAssociations.get(ontology), id,
						associationTypes);
			}
			res.add(assocs);
		}
		return res;
	}

	@Override
	public void storeAssociations(Collection<Association> associations) {
		for (Association assoc : associations) {
			storeDirectedAssociation(outboundAssociations, assoc.getSource(), assoc);
			storeDirectedAssociation(inboundAssociations, assoc.getTarget(), assoc);
		}
	}

	@Override
	public void deleteAssociations(Collection<Association> associations) {
		for (Association assoc : associations) {
			deleteDirectedAssociation(outboundAssociations, assoc.getSource(),
					assoc);
			deleteDirectedAssociation(inboundAssociations, assoc.getTarget(),
					assoc);
		}
	}

	@Override
	public Map<String, Collection<NodeDescriptor>> queryDictionary(
			Set<String> keys) {
		Map<String, Collection<NodeDescriptor>> res = new HashMap();
		for (String key : keys) {
			Set<NodeHandle> handles = dictionary.get(key);
			if (handles != null) {
				Collection<NodeDescriptor> descriptors = new ArrayList();
				res.put(key, descriptors);
				for (NodeHandle node : handles) {
					// Fetch the id and the corresponding ontology of the given
					// descriptor
					int id = node.getId();
					OntologyHandle ontology = node.getOntology();

					Map<Integer, NodeDescriptor> idMap = nodes.get(ontology);
					if (idMap != null && idMap.containsKey(id)) {
						descriptors.add(idMap.get(id));
					}
				}
			}
		}
		return res;
	}

	@Override
	public void storeDictionaryEntries(Map<NodeHandle, Set<String>> keys) {
		for (NodeHandle node : keys.keySet()) {
			Set<String> nodeKeys = dictionaryNodeIndex.get(node);
			if (nodeKeys == null) {
				nodeKeys = new HashSet();
				dictionaryNodeIndex.put(node, nodeKeys);
			}

			for (String key : keys.get(node)) {
				// Store the key -> nodes relation
				Set<NodeHandle> handles = dictionary.get(key);
				if (handles == null) {
					handles = new HashSet();
					dictionary.put(key, handles);
				}
				handles.add(node);

				// Store the node -> keys relation
				nodeKeys.add(key);
			}
		}
	}

	@Override
	public void deleteDictionaryEntries(Collection<NodeHandle> nodes) {
		for (NodeHandle node : nodes) {
			Set<String> keys = dictionaryNodeIndex.get(node);
			if (keys != null) {
				for (String key : keys) {
					Set<NodeHandle> handles = dictionary.get(key);
					handles.remove(node);
					if (handles.isEmpty()) {
						dictionary.remove(key);
					}
				}
				dictionaryNodeIndex.remove(node);
			}
		}
	}

	@Override
	public void clear() {
		ontologies.clear();
		dictionary.clear();
		dictionaryNodeIndex.clear();
		inboundAssociations.clear();
		outboundAssociations.clear();
		nodes.clear();
		rootNodes.clear();
	}

}
