/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver.implementation;

import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriverException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.HashMap;
import java.util.Map;

/**
 * The SQLStatements class bundles the possible SQL statments that can be made
 * on the OntologyDatabase. It is capable of returning a prepared statement for
 * each statement type and caches these. It contains a helper function for each
 * statements that fills in the parameters as needed.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public abstract class SQLStatements {

	/**
	 * Enum that contains IDS for all possible statements (statement
	 * identifiers, SID)
	 */
	public enum SID {

		getOntologies,
		getOntologyByName,
		getOntologyByID,
		getOntologyDate,
		setOntologyDate,
		insertOntology,
		deleteOntology,
		deleteNodesForOntology,
		deleteNodeNamesForOntology,
		deleteAssociationsForOntology,
		deleteDictionaryEntriesForOntology,
		getRootNodes,
		getIsRootNode,
		getNodeNames,
		getNodeExists,
		insertNodeID,
		deleteNodeID,
		insertNodeName,
		deleteNodeNames,
		getInboundAssociations,
		getOutboundAssociations,
		getUndirectedAssociations,
		insertAssociation,
		deleteAssociation,
		deleteNodeAssociations,
		queryDictionary,
		insertDictionaryEntry,
		deleteDictionaryEntry,
		deleteTableIfExists
	}

	/**
	 * Reference to the JDBC connection.
	 */
	private final Connection connection;

	/**
	 * Caches used for storing the prepared statments.
	 */
	private final Map<SID, PreparedStatement> statementCache;

	/**
	 * Initializes the JDBCSQLStatments class.
	 *
	 * @param connection is the database connection for which the statments
	 * should be generated.
	 */
	public SQLStatements(Connection connection) {
		this.connection = connection;
		this.statementCache = new HashMap();
	}

	/**
	 * Returns an array containg the names of each table entry.
	 *
	 * @return an array containg the names of each table entry.
	 */
	public abstract String[] getTables();

	/**
	 * Returns the an SQL string that can be used to drop a table with the given
	 * name.
	 *
	 * @param table is the table that should be droped.
	 * @return a sql string that -- if executed -- drops the given table.
	 */
	public abstract String getSQLDropTableIfExists(String table);

	/**
	 * Returns a number of SQL statments that create the schema for the given
	 * table.
	 *
	 * @param table is the name of the table for which the schema SQL should be
	 * returned.
	 * @return an array containg the statments that need to be executed.
	 */
	public abstract String[] getSchemaSQLForTable(String table);

	/**
	 * Returns the SQL for the corresponding statement id.
	 *
	 * @param statementID is the statement identifier.
	 * @return the SQL for the corresponding statement id.
	 */
	public abstract String getStatementSQL(SID statementID);

	/**
	 * Returns a prepared statement for the given SID.
	 *
	 * @param statementID is the statement identifier.
	 * @return a prepared statement for the given SID.
	 */
	public PreparedStatement getStatement(SID statementID) {
		// If the statement with the given SID already resides in the cache,
		// return it from there.
		if (statementCache.containsKey(statementID)) {
			return statementCache.get(statementID);
		}

		// Fetch the SQL for the statment
		String SQL = getStatementSQL(statementID);
		try {
			// Create the corresponding prepared statment.
			PreparedStatement stmt = connection.prepareStatement(SQL);
			statementCache.put(statementID, stmt);
			return stmt;
		} catch (SQLException ex) {
			throw new OntologyDatabaseDriverException("SQLException while creating prepared statment\n"
					+ "SQL: " + SQL + "\n"
					+ "Message: " + ex.getMessage());
		}
	}

	/**
	 * Closes the created prepared statements.
	 */
	public void finalizeStatements() {
		for (PreparedStatement stmt : statementCache.values()) {
			try {
				stmt.close();
			} catch (SQLException ex) {
				// Nothing to be done here! We wanted to close the statement anyways...
			}
		}
		statementCache.clear();
	}

	/**
	 * Returns the statement "getOntologies".
	 *
	 * @return the statement "getOntologies".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetOntologies() throws SQLException {
		return getStatement(SID.getOntologies);
	}

	/**
	 * Returns the statment "getOntologyByName".
	 *
	 * @param name is the name that is queried for.
	 * @return the statement "getOntologyByName".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetOntologyByName(String name) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getOntologyByName);
		stmt.setString(1, name);
		return stmt;
	}

	/**
	 * Returns the statement "getOntologyByID".
	 *
	 * @param oid is the ontology id.
	 * @return the statement "getOntologyByID".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetOntologyByID(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getOntologyByID);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "insertOntology".
	 *
	 * @param name is the name of the new ontology.
	 * @return the statement "insertOntology".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementInsertOntology(String name) throws SQLException {
		PreparedStatement stmt = getStatement(SID.insertOntology);
		stmt.setString(1, name);
		return stmt;
	}

	/**
	 * Returns the statement "getOntologyDate".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "getOntologyDate".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetOntologyDate(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getOntologyDate);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "getOntologyDate".
	 *
	 * @param oid is the id of the ontology.
	 * @param date is the date the ontology date should be set to.
	 * @return the statement "getOntologyDate".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementSetOntologyDate(int oid, Timestamp date) throws SQLException {
		PreparedStatement stmt = getStatement(SID.setOntologyDate);
		stmt.setTimestamp(1, date);
		stmt.setInt(2, oid);
		return stmt;
	}

	/**
	 * Returns the statement "deleteNodesForOntology".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "deleteNodesForOntology".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteOntology(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteOntology);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "deleteNodesForOntology".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "deleteNodesForOntology".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteNodesForOntology(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteNodesForOntology);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "deleteNodeNamesForOntology".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "deleteNodeNamesForOntology".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteNodeNamesForOntology(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteNodeNamesForOntology);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "deleteAssociationsForOntology".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "deleteAssociationsForOntology".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteAssociationsForOntology(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteAssociationsForOntology);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "deleteDictionaryEntriesForOntology".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "deleteDictionaryEntriesForOntology".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteDictionaryEntriesForOntology(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteDictionaryEntriesForOntology);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "getRootNodes".
	 *
	 * @param oid is the id of the ontology.
	 * @return the statement "getRootNodes".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetRootNodes(int oid) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getRootNodes);
		stmt.setInt(1, oid);
		return stmt;
	}

	/**
	 * Returns the statement "getIsRootNode".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "getIsRootNode".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetIsRootNode(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getIsRootNode);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		return stmt;
	}

	/**
	 * Returns the statement "getNodeNames".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "getNodeNames".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetNodeNames(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getNodeNames);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		return stmt;
	}

	/**
	 * Returns the statement "getNodeExists".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "getNodeExists".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetNodeExists(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getNodeExists);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		return stmt;
	}

	/**
	 * Returns the statement "insertNodeID".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @param root specifies whether the node is a root node or not.
	 * @return the statement "insertNodeID".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementInsertNodeID(int oid, int id, boolean root) throws SQLException {
		PreparedStatement stmt = getStatement(SID.insertNodeID);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		stmt.setBoolean(3, root);
		return stmt;
	}

	/**
	 * Returns the statement "deleteNodeID".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "deleteNodeID".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteNodeID(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteNodeID);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		return stmt;
	}

	/**
	 * Returns the statement "insertNodeName".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @param name is the name of the node.
	 * @param type is the type of the node.
	 * @param mode is the comparison mode.
	 * @return the statement "insertNodeName".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementInsertNodeName(int oid, int id,
			String name, String type, int mode, boolean primary) throws SQLException {
		PreparedStatement stmt = getStatement(SID.insertNodeName);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		stmt.setString(3, name);
		stmt.setString(4, type);
		stmt.setInt(5, mode);
		stmt.setBoolean(6, primary);
		return stmt;
	}

	/**
	 * Returns the statement "deleteNodeNames".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "deleteNodeNames".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteNodeNames(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteNodeNames);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		return stmt;
	}

	/**
	 * Returns the statement "getInboundAssociations".
	 *
	 * @param oid is the id of the ontology.
	 * @param tar is the id of the target node.
	 * @param type is the association type.
	 * @return the statement "getInboundAssociations".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetInboundAssociations(int oid, int tar, int type) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getInboundAssociations);
		stmt.setInt(1, oid);
		stmt.setInt(2, tar);
		stmt.setInt(3, type);
		return stmt;
	}

	/**
	 * Returns the statement "getOutboundAssociations".
	 *
	 * @param oid is the id of the ontology.
	 * @param src is the id of the source node.
	 * @param type is the association type.
	 * @return the statement "getOutboundAssociations".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetOutboundAssociations(int oid, int src, int type) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getOutboundAssociations);
		stmt.setInt(1, oid);
		stmt.setInt(2, src);
		stmt.setInt(3, type);
		return stmt;
	}

	/**
	 * Returns the statement "getUndirectedAssociations".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @param type is the association type.
	 * @return the statement "getUndirectedAssociations".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementGetUndirectedAssociations(int oid, int id, int type) throws SQLException {
		PreparedStatement stmt = getStatement(SID.getUndirectedAssociations);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		stmt.setInt(3, id);
		stmt.setInt(4, type);
		return stmt;
	}

	/**
	 * Returns the statement "insertAssociation".
	 *
	 * @param oid is the id of the ontology.
	 * @param src is the id of the source node.
	 * @param tar is the id of the target node.
	 * @param type is the association type.
	 * @return the statement "insertAssociation".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementInsertAssociation(int oid, int src, int tar, int type) throws SQLException {
		PreparedStatement stmt = getStatement(SID.insertAssociation);
		stmt.setInt(1, oid);
		stmt.setInt(2, src);
		stmt.setInt(3, tar);
		stmt.setInt(4, type);
		return stmt;
	}

	/**
	 * Returns the statement "deleteAssociation".
	 *
	 * @param oid is the id of the ontology.
	 * @param src is the id of the source node.
	 * @param tar is the id of the target node.
	 * @return the statement "deleteAssociation".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteAssociation(int oid, int src, int tar) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteAssociation);
		stmt.setInt(1, oid);
		stmt.setInt(2, src);
		stmt.setInt(3, tar);
		return stmt;
	}

	/**
	 * Returns the statement "deleteNodeAssociations".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "deleteNodeAssociations".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteNodeAssociations(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteNodeAssociations);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		stmt.setInt(3, id);
		return stmt;
	}

	/**
	 * Returns the statement "queryDictionary".
	 *
	 * @param entry is the dictionary entry that is queried.
	 * @return the statement "queryDictionary".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementQueryDictionary(String entry) throws SQLException {
		PreparedStatement stmt = getStatement(SID.queryDictionary);
		stmt.setString(1, entry);
		return stmt;
	}

	/**
	 * Returns the statement "insertDictionaryEntry".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @param entry is the dictionary entry that should be inserted.
	 * @return the statement "insertDictionaryEntry".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementInsertDictionaryEntry(int oid, int id, String entry) throws SQLException {
		PreparedStatement stmt = getStatement(SID.insertDictionaryEntry);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		stmt.setString(3, entry);
		return stmt;
	}

	/**
	 * Returns the statement "deleteDictionaryEntry".
	 *
	 * @param oid is the id of the ontology.
	 * @param id is the id of the node.
	 * @return the statement "deleteDictionaryEntry".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteDictionaryEntry(int oid, int id) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteDictionaryEntry);
		stmt.setInt(1, oid);
		stmt.setInt(2, id);
		return stmt;
	}

	/**
	 * Returns the statement "deleteTableIfExists".
	 *
	 * @param table is the name of the table that should be deleted.
	 * @return the statement "deleteTableIfExists".
	 * @throws java.sql.SQLException
	 */
	public PreparedStatement getStatementDeleteTableIfExists(String table) throws SQLException {
		PreparedStatement stmt = getStatement(SID.deleteTableIfExists);
		stmt.setString(1, table);
		return stmt;
	}

}
