/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.importer;

import java.io.File;

import de.unibi.techfak.scie.db.DatabaseSingleton;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.importer.drug_bank.DrugBankImporter;
import de.unibi.techfak.scie.ontology.importer.mesh.MeshImporter;
import de.unibi.techfak.scie.ontology.importer.mesh.MeshImporter2;
import de.unibi.techfak.scie.ontology.importer.ncbi.NCBIImporter;
import de.unibi.techfak.scie.ontology.importer.small.SmallImporter;

public class XMLImporter {

	private static final String usage
			= "Usage: XMLImporter <type> <config> <input> [override]\n"
			+ "Where <type> is the type of XMLImporter"
			+ "[ncbi|drugbank|small] <config> refers to the config.xml that "
			+ "should be used, <input> to the taxonomy input file. "
			+ "If \"override\" is specified as a fourth parameter, the "
			+ "existing ontology will be ereased and replaced, "
			+ "otherwise the program will abort if the ontology already "
			+ "exists.";

	public static void main(String[] args) {
		// Print the usage of the main program
		if (args.length != 3 && args.length != 4) {
			System.out.println(usage);
			return;
		}

		// Try to initialize the database
		DatabaseSingleton.initialize(args[1]);
		
		// Make sure the given file exists
		File sourceFile = new File(args[2]);
		if (!sourceFile.exists()) {
			System.out.println("Given file does not exist. Aborting");
			return;
		}
		
		AbstractImporter importer;
		try (OntologyDatabase db = DatabaseSingleton.getInstance().getOntologyDatabase()) {
			switch (args[0]) {
				case "ncbi":
					importer = new NCBIImporter(db);
					break;
				case "drugbank":
					importer = new DrugBankImporter(db);
					break;
				case "small":
					String ontoName = sourceFile.getName().replace(".xml", "");
					importer = new SmallImporter(db, ontoName);
					break;
				case "mesh":
					importer = new MeshImporter(db);
					break;
				case "mesh2":
					importer = new MeshImporter2(db);
					break;	
				default:
					System.out.println(usage);
					return;
			}

			// Check whether the ontology already exists, if yes abort unless
			// "override" was given as a third argument
			boolean override = (args.length == 4) && args[3].equals("override");
			System.out.println("Starting import of " + importer.ontologyName + " database from " + args[2]);
			importer.importOntology(sourceFile, override);
			System.out.println("Done.");
		}
	}
}
