/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.importer.drug_bank;

import java.util.LinkedList;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.ontology.importer.AbstractImporter;

/**
 *
 *
 */
public class DrugBankImporter extends AbstractImporter {

	/**
	 * Instanciates the importer for the given database. Uses "drug" as ontology
	 * name.
	 *
	 * @param db is a reference to the database in which the ontology should be
	 * stored.
	 */
	public DrugBankImporter(OntologyDatabase db) {
		super(db, "drug");
	}

	/**
	 * Flushes all data from Drugbank xml into db.
	 *
	 * @param onto refers to the Ontology into which the datbase should be
	 * imported.
	 * @param parser the XMLStreamReader for the Drugbank xml
	 * @throws XMLStreamException
	 */
	@Override
	protected void parseXML(OntologyHandle onto, XMLStreamReader parser) throws XMLStreamException {
		boolean nameOfDrug = false;
		boolean propertyKindFormula = false;
		int imported = 0;
		int drugOpend = 0;
		// Current node information
		int taxonId = -1;
		List<NodeName> taxonNames = new LinkedList();
		while (parser.hasNext()) {
			switch (parser.getEventType()) {

				case XMLStreamConstants.END_DOCUMENT:
					parser.close();
					break;

				case XMLStreamConstants.START_ELEMENT:
					if (parser.getLocalName().equals("name")) {
						if (nameOfDrug) {
							final String text = parser.getElementText();
							final String type = "common name";
							//System.out.println("Name: " + text);
							taxonNames.add(new NodeName(text, type, true));
							nameOfDrug = false;
						}
					} else if (parser.getLocalName().equals("brand")) {
//						final String type = "brand";
//						final String text = parser.getElementText();
//						taxonNames.add(new NodeName(text, type));
						//System.out.println("Brand: " + text);
					} else if (parser.getLocalName().equals("synonym")) {
						final String type = "synonym";
						final String text = parser.getElementText();
						if(text.length()>255){
							System.out.println("Synonym to long(Id:"+taxonId+"):\n"+text);
						}else{
							taxonNames.add(new NodeName(text, type));
						}
						//System.out.println("Synomnym: " + text);
					} else if (parser.getLocalName().equals("drug")) {
						nameOfDrug = true;
						drugOpend++;
					} else if (parser.getLocalName().equals("kind")) {
						if (parser.getElementText().equals("Molecular Formula")) {
							propertyKindFormula = true;
						}
					} else if (parser.getLocalName().equals("value") && propertyKindFormula) {
						final String text = parser.getElementText();
						final String type = "Molecular Formula";
						taxonNames.add(new NodeName(text, type, ComparisonMode.CLEAN));
						//System.out.println("Formula: " + text);
					} else if (parser.getLocalName().equals("drugbank-id")) {
						final String text = parser.getElementText();
						try {
							taxonId = new Integer(text.substring(2));
						} catch (NumberFormatException ex) {
							taxonId = -1;
						}
						//System.out.println("ID: " + text);

					}
					break;
				case XMLStreamConstants.END_ELEMENT:
					if (parser.getLocalName().equals("drug")) {
						nameOfDrug = false;
						drugOpend--;
						if (drugOpend == 0) {
							if (taxonId >= 0 && taxonNames.size() > 0) {
								Node node = new Node(db, onto, taxonId, taxonNames);
								db.storeNode(node);
							}
							taxonId = -1;
							taxonNames.clear();
							imported++;
							if (imported % 1000 == 0) {
								System.out.println(imported + " Drugs imported");
							}
						}
					} else if (parser.getLocalName().equals("property")) {
						propertyKindFormula = false;
					}
				default:
					break;
			}
			parser.next();
		}
	}
}
