/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * This class provides static methods that help to compose annotations to more
 * complex annotations in a probabilistic way.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class AnnotationCompositionUtil {

	/**
	 * This calculates the AnnotationCandidate with the maximum probability to
	 * belong to some given core for a composit annotation and returns it with
	 * the probability (without normalization!) attached.
	 *
	 * The probability is defined as the product of the syntactic and the
	 * semantic probability of those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by
	 * implementing a SemanticProbabilityModel for the
	 * AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations
	 * in the tests. Details are described in the documentation of the
	 * calculateSyntacticProbability-method.
	 *
	 * This uses the CharBasedSyntacticProbabilityModel per default.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param semanticProbModel a model that is able to calculate the
	 * probability of two annotation candidates belonging together.
	 * @param candidates an a priori given list of candidates that might belong
	 * to the core annotation.
	 * @param syntacticSigma this influences how the syntactic probability is
	 * calculated. For an in-depth documentation have a look at the
	 * calculateSyntacticProbability-method.
	 * @return returns the most probable candidate with the probability
	 * attached.
	 */
	public static <X extends AnnotationCandidate> ProbableHypothesis<X> calculateMaxProbHypothesis(
			final AnnotationCandidate core,
			final SemanticProbabilityModel<X> semanticProbModel,
			final Collection<X> candidates,
			final double syntacticSigma) {
		final SyntacticProbabilityModel syntacticProbModel =
				new CharBasedSyntacticProbabilityModel(syntacticSigma);
		return calculateMaxProbHypothesis(core, syntacticProbModel,
				semanticProbModel, candidates);
	}

	/**
	 * This calculates the AnnotationCandidate with the maximum probability to
	 * belong to some given core for a composit annotation and returns it with
	 * the probability (without normalization!) attached.
	 *
	 * The probability is defined as the product of the syntactic and the
	 * semantic probability of those two annotations belonging together.
	 *
	 * The syntactic probability should be given by the user of this function by
	 * implementing a SyntacticProbabilityModel.
	 *
	 * The semantic probability should be given by the user of this function by
	 * implementing a SemanticProbabilityModel for the
	 * AnnotationCandidate-Class.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param syntacticProbModel a model that is able to calculate the
	 * probabiliy of two AnnotationCandidates belonging together from a
	 * syntactic perspective.
	 * @param semanticProbModel a model that is able to calculate the
	 * probability of two annotation candidates belonging together from a
	 * semantic perspective.
	 * @param candidates an a priori given list of candidates that might belong
	 * to the core annotation.
	 * @return returns the most probable candidate with the probability
	 * attached.
	 */
	public static <X extends AnnotationCandidate> ProbableHypothesis<X> calculateMaxProbHypothesis(
			final AnnotationCandidate core,
			final SyntacticProbabilityModel syntacticProbModel,
			final SemanticProbabilityModel<X> semanticProbModel,
			final Collection<X> candidates) {

		//select all candidates in the range of three standard deviances. Everything else is
		//approximated by a probability of zero.
		final int effectiveRange = syntacticProbModel.getMaximumCharBasedRange();
		final int minBegin = core.getBegin() - effectiveRange;
		final int maxEnd = core.getEnd() + effectiveRange;

		//search for the most probable candidate.
		double maxAposteriori = 0;
		X map = null;
		for (final X candidate : candidates) {
			if (candidate.getBegin() < minBegin || candidate.getEnd() > maxEnd) {
				continue;
			}

			double prior = syntacticProbModel.
					calculateSyntacticProbability(core, candidate);
			double likelihood = semanticProbModel.calculateSemanticProbability(
					candidate);
			if (likelihood > 0) {
				double posterior = likelihood * prior;
				if (posterior > maxAposteriori) {
					maxAposteriori = posterior;
					map = candidate;
				}
			}
		}
		return new ProbableHypothesis<>(map, maxAposteriori);
	}

	/**
	 * For any of the given candidates the probability is calculated, that this
	 * candidate belongs to the given core annotation. The probability is
	 * defined as the product of the syntactic and the semantic probability of
	 * those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by
	 * implementing a SemanticProbabilityModel for the
	 * AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations
	 * in the tests. Details are described in the documentation of the
	 * calculateSyntacticProbability-method.
	 *
	 * This uses the CharBasedSyntacticProbabilityModel per default.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param semanticProbModel a model that is able to calculate the
	 * probability of two annotation candidates belonging together.
	 * @param candidates an a priori given list of candidates that might belong
	 * to the core annotation.
	 * @param syntacticSigma this influences how the syntactic probability is
	 * calculated. For an in-depth documentation have a look at the
	 * calculateSyntacticProbability-method.
	 * @return A subset of the given candidates sorted w.r.t. their probability.
	 * The most probable hypothesis comes first.
	 */
	public static <X extends AnnotationCandidate> ArrayList<ProbableHypothesis<X>> calculateProbableHypotheses(
			final AnnotationCandidate core,
			final SemanticProbabilityModel<X> semanticProbModel,
			final Collection<X> candidates, final double syntacticSigma) {
		final SyntacticProbabilityModel syntacticProbModel =
				new CharBasedSyntacticProbabilityModel(syntacticSigma);
		return calculateProbableHypotheses(core, syntacticProbModel,
				semanticProbModel, candidates);
	}

	/**
	 * For any of the given candidates the probability is calculated, that this
	 * candidate belongs to the given core annotation. The probability is
	 * defined as the product of the syntactic and the semantic probability of
	 * those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by
	 * implementing a SemanticProbabilityModel for the
	 * AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations
	 * in the tests. Details are described in the documentation of the
	 * calculateSyntacticProbability-method.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param syntacticProbModel a model that is able to calculate the
	 * probabiliy of two AnnotationCandidates belonging together from a
	 * syntactic perspective.
	 * @param semanticProbModel a model that is able to calculate the
	 * probability of two annotation candidates belonging together from a
	 * semantic perspective.
	 * @param candidates an a priori given list of candidates that might belong
	 * to the core annotation.
	 * @param syntacticSigma this influences how the syntactic probability is
	 * calculated. For an in-depth documentation have a look at the
	 * calculateSyntacticProbability-method.
	 * @return A subset of the given candidates sorted w.r.t. their probability.
	 * The most probable hypothesis comes first.
	 */
	public static <X extends AnnotationCandidate> ArrayList<ProbableHypothesis<X>> calculateProbableHypotheses(
			final AnnotationCandidate core,
			final SyntacticProbabilityModel syntacticProbModel,
			final SemanticProbabilityModel<X> semanticProbModel,
			final Collection<X> candidates) {

		//select all candidates in the range of three standard deviances. Everything else is
		//approximated by a probability of zero.
		final int effectiveRange = syntacticProbModel.getMaximumCharBasedRange();
		final int minBegin = core.getBegin() - effectiveRange;
		final int maxEnd = core.getEnd() + effectiveRange;

		//calculate the probabilities.
		final ArrayList<ProbableHypothesis<X>> hypotheses = new ArrayList<>();

		for (final X candidate : candidates) {
			if (candidate.getBegin() < minBegin || candidate.getEnd() > maxEnd) {
				continue;
			}

			double syntacticProb = syntacticProbModel.
					calculateSyntacticProbability(core, candidate);
			double semanticProb = semanticProbModel.
					calculateSemanticProbability(
					candidate);
			if (semanticProb > 0) {
				double prob = semanticProb * syntacticProb;
				hypotheses.add(new ProbableHypothesis<>(candidate, prob));
			}
		}
		//sort the hypotheses according to their probability
		Collections.sort(hypotheses);
		//and we want the most likely hypothesis first
		Collections.reverse(hypotheses);
		return hypotheses;
	}

	/**
	 * This calculates all composita that are possible to create given a certain
	 * core annotation and a list of candidates for each slots of the
	 * compositum.
	 *
	 * Please note that this method assumes that the compositum is structured as
	 * follows:
	 *
	 * <code>class CompositumClass {
	 *
	 *	CoreClass core;
	 *	Slot1Class slot1Instance;
	 *	Slot2Class slot2Instance;
	 *	...
	 * }</code>
	 *
	 * The user is required to provide a SemanticProbabilityModel and a
	 * collection of a priori candidates for each slot.
	 *
	 * Also she is required to construct a class that is able to actually create
	 * the compositum from one candidate for each slot.
	 *
	 * This uses the CharBasedSyntacticProbabilityModel per default.
	 *
	 * @param core the core annotation that shall be the center of the new
	 * compositum.
	 * @param semanticModels An array of SemanticProbabilityModels, one for each
	 * slot.
	 * @param candidates An Array of Candidate-collections, one for each slot in
	 * the same order as the SemanticProbabilityModels.
	 * @param compositor An AnnotationCompositor instance to construct the
	 * resulting composita from all possible hypothesis combinations.
	 * @param syntacticSigma this influences how the syntactic probability is
	 * calculated. For an in-depth documentation have a look at the
	 * calculateSyntacticProbability-method.
	 * @return a collection of constructed composita as it is done by the given
	 * compositor.
	 */
	public static <Y> List<Y> calculateComposita(
			final AnnotationCandidate core,
			final SemanticProbabilityModel[] semanticModels,
			final Collection[] candidates,
			final AnnotationCompositor<Y> compositor,
			final double syntacticSigma) {
		final SyntacticProbabilityModel syntacticProbModel =
				new CharBasedSyntacticProbabilityModel(syntacticSigma);
		return calculateComposita(core, syntacticProbModel, semanticModels,
				candidates, compositor);
	}

	/**
	 * This calculates all composita that are possible to create given a certain
	 * core annotation and a list of candidates for each slots of the
	 * compositum.
	 *
	 * Please note that this method assumes that the compositum is structured as
	 * follows:
	 *
	 * <code>class CompositumClass {
	 *
	 *	CoreClass core;
	 *	Slot1Class slot1Instance;
	 *	Slot2Class slot2Instance;
	 *	...
	 * }</code>
	 *
	 * The user is required to provide a SemanticProbabilityModel and a
	 * collection of a priori candidates for each slot.
	 *
	 * Also she is required to construct a class that is able to actually create
	 * the compositum from one candidate for each slot.
	 *
	 * @param core the core annotation that shall be the center of the new
	 * compositum.
	 * @param syntacticModel a model that is able to calculate the probabiliy of
	 * two AnnotationCandidates belonging together from a syntactic perspective.
	 * @param semanticModels An array of SemanticProbabilityModels, one for each
	 * slot.
	 * @param candidates An Array of Candidate-collections, one for each slot in
	 * the same order as the SemanticProbabilityModels.
	 * @param compositor An AnnotationCompositor instance to construct the
	 * resulting composita from all possible hypothesis combinations.
	 * @return a collection of constructed composita as it is done by the given
	 * compositor.
	 */
	public static <Y> List<Y> calculateComposita(
			final AnnotationCandidate core,
			final SyntacticProbabilityModel syntacticModel,
			final SemanticProbabilityModel[] semanticModels,
			final Collection[] candidates,
			final AnnotationCompositor<Y> compositor) {
		//check input validity.
		if (semanticModels.length != candidates.length) {
			throw new RuntimeException(
					"We need one SemanticProbabilityModel for each candidate collection!");
		}

		//get all possible hypotheses for the single slots.
		final ArrayList<ArrayList<ProbableHypothesis>> slotHypos =
				new ArrayList<>(candidates.length);
		for (int i = 0; i < candidates.length; i++) {
			final ArrayList<ProbableHypothesis> currentSlotHypos =
					calculateProbableHypotheses(core, syntacticModel,
					semanticModels[i], candidates[i]);
			slotHypos.add(currentSlotHypos);
		}
		//if we have all those lists recursively create all possible hypotheses.
		ArrayList<Y> composita = new ArrayList<>();
		calculateCompositaRecursion(core, slotHypos,
				new ArrayList<ProbableHypothesis>(), composita, compositor, 0);
		return composita;
	}

	private static <Y> void calculateCompositaRecursion(
			final AnnotationCandidate core,
			final ArrayList<ArrayList<ProbableHypothesis>> slotHypos,
			ArrayList<ProbableHypothesis> currentHypoCombination,
			final Collection<Y> composita,
			final AnnotationCompositor<Y> compositor,
			int depth) {
		/* 
		 * if we have gone through all lists we create the actual compositon
		 * and add it to the composita collection.
		 */
		if (depth == slotHypos.size()) {
			final Y currentComposition = compositor.compose(core,
					currentHypoCombination);
			if (currentComposition != null) {
				composita.add(currentComposition);
			}
			return;
		}
		//otherwise get all hypotheses for the current slot.
		if (slotHypos.get(depth).isEmpty()) {
			//if there are none, go into recursion with null at this slot.
			currentHypoCombination.add(null);
			calculateCompositaRecursion(core, slotHypos, currentHypoCombination,
					composita, compositor, depth + 1);
		} else {
			for (ProbableHypothesis slotHypo : slotHypos.get(depth)) {
				//go into recursion for each possibility
				final ArrayList<ProbableHypothesis> newCombination =
						new ArrayList<>(currentHypoCombination);
				newCombination.add(slotHypo);
				calculateCompositaRecursion(core, slotHypos, newCombination,
						composita, compositor, depth + 1);
			}
		}
	}
}
