/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model;

/**
 * This is an interface for classes that are able to calculate the probability
 * that two given AnnotationCandidates belong together from a syntactic
 * perspective.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public interface SyntacticProbabilityModel {

	/**
	 * Returns the probability that two given AnnotationCandidates belong
	 * together from a syntactic perspective.
	 *
	 * The implementing class garuantees that certain critria are met by this
	 * implementation:
	 * <ul><li>The returned value is between 0 and 1.</li>
	 * <li>The returned value is related to the distance of both candidates in
	 * the text. If they are further apart the value should go down.</li>
	 * <li>The self-probability calculateSyntacticProbability(a,a) should be 1
	 * or close to 1.</li>
	 * <li>The probability should converge against zero fast for annotations
	 * that are far away.</li>
	 * <li>The method should be symmetric, meaning that
	 * calculateSyntacticProbability(a,b) =
	 * calculateSyntacticProbability(b,a)</li></ul>
	 *
	 * @param a the first candidate.
	 * @param b the second candidate.
	 * @return the probability that two given AnnotationCandidates belong
	 * together from a syntactic perspective.
	 */
	public double calculateSyntacticProbability(AnnotationCandidate a,
			AnnotationCandidate b);

	/**
	 * Returns a heuristic approximation which region of the text should be
	 * considered at all from a syntactic perspective. How many characters of
	 * range should be allowed in each direction?
	 *
	 * @return a heuristic approximation which region of the text should be
	 * considered at all from a syntactic perspective. How many characters of
	 * range should be allowed in each direction?
	 */
	public int getMaximumCharBasedRange();
}
