/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.ontology.DictionaryMatch;
import de.unibi.techfak.scie.ontology.Dictionary;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import java.util.ArrayList;
import java.util.List;
import static junit.framework.Assert.assertEquals;
import junit.framework.TestCase;

/**
 *
 * @author andreas
 */
public class DictionaryTest extends TestCase {

	private class TestResult {

		public final String word;
		public final double probability;

		public TestResult(String word, double probability) {
			this.word = word;
			this.probability = probability;
		}
	}

	private void checkMatches(List<DictionaryMatch> matches,
			TestResult[] results) {
		// Compare the count of matches
		assertEquals("Matches length", results.length, matches.size());

		// Compare the test results with the actual match list
		for (TestResult result : results) {
			// Try to search the given result in the match list
			int matchIdx = -1;
			for (int i = 0; i < matches.size(); i++) {
				final DictionaryMatch match = matches.get(i);
				final int b = match.getBegin();
				final int e = match.getEnd();
				final String w = match.getNameDescr().getName().substring(b, e);
				final double pDiff = match.getProbability() - result.probability;
				final double eps = 0.00001;
				if (w.equals(result.word) && Math.abs(pDiff) < eps) {
					matchIdx = i;
					break;
				}
			}

			// Make sure it was really found
			assertTrue("Match " + result.word + " with probability "
					+ result.probability + " found", matchIdx >= 0);

			// Delete this match
			matches.remove(matchIdx);
		}
	}

	public DictionaryTest(String testName) {
		super(testName);
	}

	/**
	 * Test of fillMatchesForNode method, of class Dictionary.
	 */
	public void testFillMatchesForNode() {
		System.out.println("fillMatchesForNode");

		// Iterate over all comparison modes
		for (ComparisonMode mode : ComparisonMode.values()) {
			// Gerate a test name and a test node descriptor
			NodeName name = new NodeName(
					"what is7 an ocean, but a multitude of Drops?",
					mode);
			NodeDescriptor node = new NodeDescriptor(null, name);

			// Search the results for some test words, check the results
			// depending on the comparison mode
			String testWords[] = new String[]{
				"Drop",
				"Drops",
				"Drops?",
				"drops?",
				"ocean",
				"ocean,",
				"is",
				"is7",
				"what",
				"but"
			};

			// Iterate over all test words, fetch all matches for the test node
			// descriptor with the test word
			ArrayList<DictionaryMatch> matches = new ArrayList();
			for (String test : testWords) {
				Dictionary.fillMatchesForNode(matches, node, test, 0.5);
			}

			// Print the matches for debugging
			System.out.println("----------");
			for (DictionaryMatch match: matches) {
				final String s = name.getName().substring(match.getBegin(), match.getEnd());
				System.out.println("Found " + s + " @ " + mode.toString());
			}

			// Check the match results for each mode individuall
			switch (mode) {
				case FUZZY:
					checkMatches(matches, new TestResult[]{
						new TestResult("Drops", 4.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("ocean", 5.0 / 5.0)
					});
					break;
				case SOFT:
					checkMatches(matches, new TestResult[]{
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("is7", 2.0 / 2.0),
						new TestResult("is7", 2.0 / 2.0),
						new TestResult("what", 4.0 / 4.0),
						new TestResult("but", 4.0 / 4.0),
					});
					break;
				case CLEAN:
					checkMatches(matches, new TestResult[]{
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("is7", 3.0 / 3.0),
						new TestResult("what", 4.0 / 4.0),
						new TestResult("but", 4.0 / 4.0),
					});
					break;
				case LOWERCASE:
					checkMatches(matches, new TestResult[]{
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 6.0 / 6.0),
						new TestResult("Drops", 6.0 / 6.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("ocean", 6.0 / 6.0),
						new TestResult("is7", 3.0 / 3.0),
						new TestResult("what", 4.0 / 4.0),
						new TestResult("but", 4.0 / 4.0),
					});
					break;
				case STRICT:
					checkMatches(matches, new TestResult[]{
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("Drops", 5.0 / 5.0),
						new TestResult("ocean", 6.0 / 6.0),
						new TestResult("ocean", 5.0 / 5.0),
						new TestResult("is7", 3.0 / 3.0),
						new TestResult("what", 4.0 / 4.0),
						new TestResult("but", 4.0 / 4.0),
					});
					break;
				case NONE:
					checkMatches(matches, new TestResult[]{});
					break;
				default:
					throw new AssertionError(mode.name());
			}
		}
	}

}
