/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model;

import de.unibi.techfak.scie.probability_model.AnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.AnnotationCandidate;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SemanticProbabilityModel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class AnnotationCompositionUtilTest extends TestCase {

	public AnnotationCompositionUtilTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	public static class TrivialAnnotationCandidate implements
			AnnotationCandidate {

		private int begin;
		private int end;
		private int value;

		public TrivialAnnotationCandidate() {
		}

		public void setBegin(int begin) {
			this.begin = begin;
		}

		public void setEnd(int end) {
			this.end = end;
		}

		public void setValue(int value) {
			this.value = value;
		}

		@Override
		public int getBegin() {
			return begin;
		}

		@Override
		public int getEnd() {
			return end;
		}

		public int getValue() {
			return value;
		}
	}

	/**
	 * Test of calculateMaxProbHypothesis method, of class
	 * AnnotationCompositionUtil.
	 */
	public void testCalculateMaxProbHypothesis() {

		final double sigma = 50;
		//we create a trivial likelihood model that only gives high probability for numbers > 100.
		final SemanticProbabilityModel<TrivialAnnotationCandidate> semanticModel =
				new SemanticProbabilityModel<TrivialAnnotationCandidate>() {
			@Override
			public double calculateSemanticProbability(
					TrivialAnnotationCandidate candidate) {
				if (candidate.getValue() > 100) {
					return 1;
				} else {
					return 0.01;
				}
			}
		};

		//then we create some candidates (even closer ones but below 100)
		final TrivialAnnotationCandidate core = new TrivialAnnotationCandidate();
		core.setBegin(0);
		core.setEnd(1);
		final TrivialAnnotationCandidate a = new TrivialAnnotationCandidate();
		a.setBegin(1);
		a.setEnd(2);
		a.setValue(1);
		final TrivialAnnotationCandidate b = new TrivialAnnotationCandidate();
		b.setBegin(2);
		b.setEnd(3);
		b.setValue(1);
		final TrivialAnnotationCandidate c = new TrivialAnnotationCandidate();
		c.setBegin(3);
		c.setEnd(4);
		c.setValue(101);
		final TrivialAnnotationCandidate d = new TrivialAnnotationCandidate();
		d.setBegin((int) sigma * 4);
		d.setEnd((int) sigma * 4 + 1);
		d.setValue(101);

		//if we calculate the most probable candidate for an empty candidate set the result should be null.
		final ArrayList<TrivialAnnotationCandidate> candidates =
				new ArrayList<>();

		ProbableHypothesis<TrivialAnnotationCandidate> maxProbHyp =
				AnnotationCompositionUtil.calculateMaxProbHypothesis(core,
				semanticModel, candidates, sigma);
		assertNull(maxProbHyp.getCandidate());
		assertEquals(0, maxProbHyp.getProbability(), 0);
		//if we only consider an annotation that is too far away, the result should be null too.
		candidates.add(d);
		maxProbHyp =
				AnnotationCompositionUtil.calculateMaxProbHypothesis(core,
				semanticModel, candidates, sigma);
		assertNull(maxProbHyp.getCandidate());
		assertEquals(0, maxProbHyp.getProbability(), 0);
		candidates.clear();

		//if we calculate the map for a and b, a should be taken because it is closer.
		candidates.add(a);
		candidates.add(b);
		maxProbHyp = AnnotationCompositionUtil.calculateMaxProbHypothesis(core,
				semanticModel, candidates, sigma);
		assertSame(a, maxProbHyp.getCandidate());
		final double lowPost = maxProbHyp.getProbability();

		//if we add c to the candidate list however, c should be taken. It is not closer, but it
		//has a value > 100.
		candidates.add(c);
		maxProbHyp = AnnotationCompositionUtil.calculateMaxProbHypothesis(core,
				semanticModel, candidates, sigma);
		assertSame(c, maxProbHyp.getCandidate());
		final double highPost = maxProbHyp.getProbability();
		//the probability should even be higher than before
		assertTrue(highPost > lowPost);
	}

	public void testCalculateProbability() {

		final double sigma = 50;
		//we create a trivial likelihood model that only gives high probability for numbers > 100.
		final SemanticProbabilityModel<TrivialAnnotationCandidate> semanticModel =
				new SemanticProbabilityModel<TrivialAnnotationCandidate>() {
			@Override
			public double calculateSemanticProbability(
					TrivialAnnotationCandidate candidate) {
				if (candidate.getValue() > 100) {
					return 1;
				} else {
					return 0.01;
				}
			}
		};

		//then we create some candidates (even closer ones but below 100)
		final TrivialAnnotationCandidate core = new TrivialAnnotationCandidate();
		core.setBegin(0);
		core.setEnd(1);
		final TrivialAnnotationCandidate a = new TrivialAnnotationCandidate();
		a.setBegin(1);
		a.setEnd(2);
		a.setValue(1);
		final TrivialAnnotationCandidate b = new TrivialAnnotationCandidate();
		b.setBegin(2);
		b.setEnd(3);
		b.setValue(1);
		final TrivialAnnotationCandidate c = new TrivialAnnotationCandidate();
		c.setBegin(3);
		c.setEnd(4);
		c.setValue(101);
		final TrivialAnnotationCandidate d = new TrivialAnnotationCandidate();
		d.setBegin((int) sigma * 4);
		d.setEnd((int) sigma * 4 + 1);
		d.setValue(101);

		//if we calculate the probability for an empty candidate set the result should be empty, too
		final ArrayList<TrivialAnnotationCandidate> candidates =
				new ArrayList<>();
		List<ProbableHypothesis<TrivialAnnotationCandidate>> resultCandidates =
				AnnotationCompositionUtil.calculateProbableHypotheses(core,
				semanticModel, candidates, sigma);
		assertNotNull(resultCandidates);
		assertTrue(resultCandidates.isEmpty());
		//now consider all candidates and check the result.
		candidates.add(a);
		candidates.add(b);
		candidates.add(c);
		candidates.add(d);
		resultCandidates =
				AnnotationCompositionUtil.calculateProbableHypotheses(core,
				semanticModel, candidates, sigma);
		assertNotNull(resultCandidates);
		//the most probable one should be c, because its value is over 100.
		final Iterator<ProbableHypothesis<TrivialAnnotationCandidate>> iterator =
				resultCandidates.iterator();
		assertTrue(iterator.hasNext());
		ProbableHypothesis<TrivialAnnotationCandidate> probCandidate = iterator.
				next();
		double prob = probCandidate.getProbability();
		assertSame(c, probCandidate.getCandidate());
		//the next one should be a, then b, because they are close.
		assertTrue(iterator.hasNext());
		probCandidate = iterator.next();
		assertSame(a, probCandidate.getCandidate());
		assertTrue(prob > probCandidate.getProbability());
		prob = probCandidate.getProbability();

		assertTrue(iterator.hasNext());
		probCandidate = iterator.next();
		assertSame(b, probCandidate.getCandidate());
		assertTrue(prob > probCandidate.getProbability());
		//d should not be part of the result list because it is too far away
		assertFalse(iterator.hasNext());
	}
}
