/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model.uima;

import de.unibi.techfak.scie.probability_model.uima.UIMAWordSyntacticProbabilityModel;
import de.unibi.techfak.scie.annotators.structure.StanfordTokenizer;
import de.unibi.techfak.isy.lppba.descriptors.Paragraph;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate;
import java.util.HashMap;
import junit.framework.TestCase;
import org.apache.uima.UIMAException;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class UIMAWordSyntacticProbabilityModelTest extends TestCase {

	public UIMAWordSyntacticProbabilityModelTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateSyntacticProbability method, of class UIMAWordSyntacticProbabilityModel.
	 */
	public void testCalculateSyntacticProbability() throws UIMAException {

		//create example
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		jcas.setDocumentText(
				"This is a test sentence with \"hold\" that gets wonderfully annotated. Also we have a second test sentence to be able to make distance tests. Even a third one.\n\nFinally we start a new paragraph.");
		//preprosess using StanfordTokenizer and ParagraphTokenizer
		final StanfordTokenizer st = new StanfordTokenizer();
		st.process(jcas);
		//insert the paragraphs artificially.
		final int midParIdx = jcas.getDocumentText().indexOf("Even a third one.") +
				"Even a third one.".length();
		Paragraph par = new Paragraph(jcas, 0, midParIdx);
		par.addToIndexes();
		par = new Paragraph(jcas, midParIdx + 2, jcas.getDocumentText().length());
		par.addToIndexes();

		//then get a mapping of words to annotations.
		final HashMap<String, Token> words = new HashMap<>();
		for (final Token word : JCasUtil.select(jcas, Token.class)) {
			words.put(word.getCoveredText(), word);
		}

		//set up the instance
		final int sigma = 6;
		final UIMAWordSyntacticProbabilityModel instance =
				new UIMAWordSyntacticProbabilityModel(sigma, jcas);

		instance.setSentenceMalus(4);
		assertEquals(4, instance.getSentenceMalus());
		instance.setParagraphMalus(8);
		assertEquals(8, instance.getParagraphMalus());

		//first we calculate the distance within the same sentence.
		final Token hold = words.get("hold");
		final Token wonderfully = words.get("wonderfully");
		assertNotNull(hold);
		assertNotNull(wonderfully);

		double result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(hold),
				new UIMAAnnotationCandidate(wonderfully));
		//There are 4 tokens in between, including the "
		double expected = instance.calculateNormalDistribution(4);
		assertEquals(expected, result, 0);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(wonderfully),
				new UIMAAnnotationCandidate(hold));
		assertEquals(expected, result, 0);

		//test for two words that are not in the same sentence.
		final Token able = words.get("able");
		assertNotNull(able);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(wonderfully),
				new UIMAAnnotationCandidate(able));
		//There are 12 tokens in between, including the .
		expected = instance.calculateNormalDistribution(12 + instance.getSentenceMalus());
		assertEquals(expected, result, 0);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(able),
				new UIMAAnnotationCandidate(wonderfully));
		assertEquals(expected, result, 0);

		//test for two words that are even two sentences away.
		final Token third = words.get("third");
		assertNotNull(third);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(wonderfully),
				new UIMAAnnotationCandidate(third));
		//There are 20 tokens in between, including two .
		expected = instance.calculateNormalDistribution(20 + 2 * instance.getSentenceMalus());
		assertEquals(expected, result, 0);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(third),
				new UIMAAnnotationCandidate(wonderfully));
		assertEquals(expected, result, 0);

		//test for two words that are not in the same paragraph.
		final Token Finally = words.get("Finally");
		assertNotNull(Finally);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(third),
				new UIMAAnnotationCandidate(Finally));
		//There are three tokens in between, including a .
		expected = instance.calculateNormalDistribution(3 +
				instance.getParagraphMalus() +
				instance.getSentenceMalus());
		assertEquals(expected, result, 0);
		result = instance.calculateSyntacticProbability(
				new UIMAAnnotationCandidate(Finally),
				new UIMAAnnotationCandidate(third));
		assertEquals(expected, result, 0);
	}

	/**
	 * Test of getMaximumCharBasedRange method, of class UIMAWordSyntacticProbabilityModel.
	 */
	public void testGetMaximumCharBasedRange() {
		final int sigma = 4;
		final UIMAWordSyntacticProbabilityModel instance =
				new UIMAWordSyntacticProbabilityModel(sigma, null);
		int expResult = sigma * 3 *
				UIMAWordSyntacticProbabilityModel.APPROXCHARSPERWORD;
		int result = instance.getMaximumCharBasedRange();
		assertEquals(expResult, result);
	}
}
