/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.importer.ncbi;

import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.AssociationType;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.ontology.importer.AbstractImporter;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

/**
 * The NCBIImporter class imports the nodes and names from the NCBI database
 * into the ontology database.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class NCBIImporter extends AbstractImporter {

	/**
	 * Instanciates the importer for the given database. Uses "ncbi" as ontology
	 * name.
	 *
	 * @param db is a reference to the database in which the ontology should be
	 * stored.
	 */
	public NCBIImporter(OntologyDatabase db) {
		super(db, "ncbi");
	}

	private enum ParserState {

		stateUnspecified(null, null),
		stateTopLevel(null, null),
		stateTaxaSet("TaxaSet", stateTopLevel),
		stateTaxon("Taxon", stateTaxaSet),
		stateOthernames("OtherNames", stateTaxon),
		stateOtherName("Name", stateOthernames);

		public final String tagname;
		public final ParserState parentState;

		private ParserState(String tagname, ParserState parentState) {
			this.tagname = tagname;
			this.parentState = parentState;
		}

	};

	private class ParserStateStack {

		private final Stack<ParserState> states;

		private ParserState currentState;

		public ParserStateStack() {
			this.states = new Stack();
			this.currentState = ParserState.stateTopLevel;
		}

		public void handleStartElement(String tagname) {
			// Check whether we can validly switch to one of the given states
			for (ParserState state : ParserState.values()) {
				if (state.parentState == currentState
						&& tagname.equals(state.tagname)) {
					states.push(currentState);
					currentState = state;
					return;
				}
			}

			// Otherwise switch to the "unspecified state"
			states.push(this.currentState);
			currentState = ParserState.stateUnspecified;
		}

		public void handleEndElement(String tagname) {
			if (!states.empty()) {
				if (tagname.equals(currentState.tagname)
						|| currentState == ParserState.stateUnspecified) {
					currentState = states.pop();
				}
			}
		}

		public ParserState getState() {
			return currentState;
		}
	}

	public static Collection<NodeName> cleanNodeNames(Collection<NodeName> names) {
		Map<String, NodeName> res = new HashMap();
		Matcher m;
		for (NodeName nodeName : names) {
			// Fetch the current name
			String name = nodeName.getName();
			m = REDUCE_SPACES.matcher(name);
			name = m.replaceAll(" ").trim();

			// Remove everything after and including "sp.", remove any reference
			// to a journal or an author
			for (Pattern reducer : CLEAN_REDUCE) {
				m = reducer.matcher(name);
				if (m.matches()) {
					name = m.group(1);
				}
			}

			// Remove all text in parentheses or in "" or in '', trim the name
			m = CLEAN_PARANTHESIS.matcher(name);
			name = m.replaceAll("").trim();

			// Create a new name and add it to the collection, merge the types
			if (!res.containsKey(name)) {
				res.put(name, new NodeName(name, nodeName.getType()));
			} else {
				NodeName oldNode = res.get(name);
				String type = oldNode.getType();
				if (!type.contains(nodeName.getType())) {
					type = type + ", " + nodeName.getType();
				}
				res.put(name, new NodeName(name, type));
			}
		}
		return res.values();
	}

	protected void parseXML(OntologyHandle onto, XMLStreamReader parser) throws XMLStreamException {
		// Class which handles the current parser state
		final ParserStateStack states = new ParserStateStack();

		// Current node information
		int taxonId = -1;
		int taxonParentId = -1;
		List<NodeName> taxonNames = new LinkedList<NodeName>();

		int cnt = 0;

		while (parser.hasNext()) {
			// Fetch the tag name
			final int eventType = parser.getEventType();
			String tagName = "";
			if (eventType == XMLStreamConstants.START_ELEMENT
					|| eventType == XMLStreamConstants.END_ELEMENT) {
				tagName = parser.getLocalName();
			}

			switch (eventType) {

				case XMLStreamConstants.END_DOCUMENT:
					parser.close();
					break;

				case XMLStreamConstants.START_ELEMENT:
					switch (states.getState()) {
						case stateTaxaSet:
							if (tagName.equals("Taxon")) {
								// Initialize all data structures upon reaching a
								// "Taxon" node and switch to the "stateTaxon" state
								taxonId = -1;
								taxonNames.clear();
							}
							break;
						case stateTaxon:
							// Parse the names and ids
							switch (tagName) {
								case "ScientificName":
									final String name = parser.getElementText();
									final String type = "scientific name";
									taxonNames.add(new NodeName(name, type));
									break;
								case "TaxId":
									try {
										taxonId = new Integer(parser.getElementText());
									} catch (NumberFormatException ex) {
										taxonId = -1;
									}
									break;
								case "ParentTaxId":
									try {
										taxonParentId = new Integer(parser.getElementText());
									} catch (NumberFormatException ex) {
										taxonParentId = -1;
									}
									break;
							}
							break;
						case stateOthernames:
							switch (tagName) {
								case "GenbankCommonName": {
									final String name = parser.getElementText();
									final String type = "genbank common name";
									taxonNames.add(new NodeName(name, type));
									break;
								}
								case "Synonym": {
									final String name = parser.getElementText();
									final String type = "synonym";
									taxonNames.add(new NodeName(name, type));
									break;
								}
								case "CommonName": {
									final String name = parser.getElementText();
									final String type = "general common name";
									taxonNames.add(new NodeName(name, type));
									break;
								}
							}
							break;
					}

					// Switch to another state if a suitable element is found
					states.handleStartElement(tagName);
					break;

				case XMLStreamConstants.END_ELEMENT:
					switch (states.getState()) {
						case stateTaxon:
							if (tagName.equals("Taxon")) {
								// Store the node in the database (if sufficient data was
								// gathered) and return to the top level.
								if (taxonId >= 0 && taxonNames.size() > 0) {
									Node node = new Node(db, onto, taxonId,
											cleanNodeNames(taxonNames));
									db.storeNode(node);

									// Store the association to the parent node
									if (taxonParentId > 0) {
										final NodeHandle parentNode
												= new NodeHandle(onto, taxonParentId);
										node.addAssociationTo(parentNode, AssociationType.IS_A);
									}

									// Display some progress information
									cnt++;
									if (cnt % 1000 == 0) {
										System.out.println("Imported " + cnt + " nodes");
									}
								}
							}
							break;
					}

					// Switch to another state if a suitable element is found
					states.handleEndElement(tagName);
					break;

				default:
					break;
			}

			// Only continue if the event type did not change over the course of
			// the processing -- otherwise iterate again in order to process the
			// new event
			if (parser.getEventType() == eventType) {
				parser.next();
			}
		}
	}

}
